"""
The tunnel community.

Author(s): Egbert Bouman
"""
from __future__ import absolute_import
from __future__ import division

from binascii import hexlify, unhexlify

from twisted.internet.task import LoopingCall

from .caches import *
from .endpoint import TunnelEndpoint
from .payload import *
from .tunnel import *
from .tunnelcrypto import CryptoException, TunnelCrypto
from ...community import Community
from ...lazy_community import lazy_wrapper, lazy_wrapper_unsigned
from ...messaging.deprecated.encoding import decode, encode
from ...messaging.payload_headers import BinMemberAuthenticationPayload
from ...peer import Peer
from ...requestcache import RequestCache
from ...util import addCallback

message_to_payload = {
    u"data": (0, DataPayload),
    u"cell": (1, CellPayload),
    u"create": (2, CreatePayload),
    u"created": (3, CreatedPayload),
    u"extend": (4, ExtendPayload),
    u"extended": (5, ExtendedPayload),
    u"ping": (6, PingPayload),
    u"pong": (7, PongPayload),
    u"destroy": (10, DestroyPayload),
    u"establish-intro": (11, EstablishIntroPayload),
    u"intro-established": (12, IntroEstablishedPayload),
    u"establish-rendezvous": (15, EstablishRendezvousPayload),
    u"rendezvous-established": (16, RendezvousEstablishedPayload),
    u"create-e2e": (17, CreateE2EPayload),
    u"created-e2e": (18, CreatedE2EPayload),
    u"link-e2e": (19, LinkE2EPayload),
    u"linked-e2e": (20, LinkedE2EPayload),
    u"peers-request": (21, PeersRequestPayload),
    u"peers-response": (22, PeersResponsePayload)
}


def tc_lazy_wrapper_unsigned(*payloads):
    """
    This function wrapper will unpack just the normal payloads for you, and handle a singular circuit_id parameter at
    the end of the parameter list

    You can now write your non-authenticated and signed functions as follows:

    ::

        @tc_lazy_wrapper_unsigned(GlobalTimeDistributionPayload, IntroductionRequestPayload,
                                  IntroductionResponsePayload, circuit_id)
        def on_message(source_address, payload1, payload2):
            '''
            :type source_address: str
            :type payload1: IntroductionRequestPayload
            :type payload2: IntroductionResponsePayload
            '''
            pass
    """
    def decorator(func):
        def wrapper(self, source_address, data, circuit_id=None):

            @lazy_wrapper_unsigned(*payloads)
            def inner_wrapper(inner_self, inner_source_address, *pyls):
                combo = list(pyls) + [circuit_id]
                return func(inner_self, inner_source_address, *combo)

            return inner_wrapper(self, source_address, data)
        return wrapper
    return decorator


class TunnelSettings(object):

    def __init__(self):
        self.crypto = TunnelCrypto()

        self.min_circuits = 1
        self.max_circuits = 8
        self.max_joined_circuits = 100

        # Maximum number of seconds that a circuit should exist
        self.max_time = 10 * 60
        # Maximum number of seconds that an introduction point should exist
        self.max_time_ip = 24 * 60 * 60
        # Maximum number of seconds before a circuit is considered inactive (and is removed)
        self.max_time_inactive = 20
        self.max_traffic = 250 * 1024 * 1024

        self.max_packets_without_reply = 50

        # Maximum number of seconds circuit creation is allowed to take. Within this time period, the unverified hop
        # of the circuit can still change in case it is unresponsive.
        self.circuit_timeout = 60
        # Maximum number of seconds that a hop allows us to change the next hop
        self.unstable_timeout = 60
        # Maximum number of seconds adding a single hop to a circuit is allowed to take.
        self.next_hop_timeout = 10

        self.swarm_lookup_interval = 30
        self.swarm_connection_limit = 5

        # We have a small delay when removing circuits/relays/exit nodes. This is to allow some post-mortem data
        # to flow over the circuit (i.e. bandwidth payouts to intermediate nodes in a circuit).
        self.remove_tunnel_delay = 5

        self.num_ip_circuits = 3
        self.peer_flags = PEER_FLAG_RELAY


class TunnelCommunity(Community):

    version = b'\x02'
    master_peer = Peer(unhexlify("4c69624e61434c504b3adbd575f1902f1d39debeb3d7b576dff1652ab6cede21c14545d134219b57b"
                                 "413a597c87c2e37b8c9f99938e0d4db3cd4a79d6b53fb5bc0cb7d222d5c6eabfaa9"))

    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings', TunnelSettings())
        self.dht_provider = kwargs.pop('dht_provider', None)
        if isinstance(self.settings, dict):
            settings = TunnelSettings()
            for k, v in self.settings.items():
                setattr(settings, k, v)
            self.settings = settings

        super(TunnelCommunity, self).__init__(*args, **kwargs)

        self.request_cache = RequestCache()

        # Messages that can arrive from the socket
        self.decode_map.update({
            chr(1): self.on_cell,
            chr(10): self.on_destroy
        })

        # Messages that can arrive from a circuit (i.e., they are wrapped in a cell)
        self.decode_map_private = {
            chr(0): self.on_data,
            chr(2): self.on_create,
            chr(3): self.on_created,
            chr(4): self.on_extend,
            chr(5): self.on_extended,
            chr(6): self.on_ping,
            chr(7): self.on_pong
        }

        self.select_index = -1
        self.circuits = {}
        self.directions = {}
        self.relay_from_to = {}
        self.relay_session_keys = {}
        self.exit_sockets = {}
        self.circuits_needed = defaultdict(int)
        self.num_hops_by_downloads = defaultdict(int)  # Keeps track of the number of hops required by downloads
        self.candidates = {}  # Keeps track of the candidates that want to be a relay/exit node

        self.crypto = self.settings.crypto

        self.logger.info("Setting exitnode = %s", self.settings.peer_flags & PEER_FLAG_EXIT_ANY)

        self.crypto.initialize(self.my_peer.key)

        if isinstance(self.endpoint, TunnelEndpoint):
            self.endpoint.set_tunnel_community(self)
            self.endpoint.set_anonymity(self._prefix, False)

        self.register_task("do_circuits", LoopingCall(self.do_circuits)).start(5, now=True)
        self.register_task("do_ping", LoopingCall(self.do_ping)).start(PING_INTERVAL)

    def unload(self):
        # Remove all circuits/relays/exitsockets
        for circuit_id in list(self.circuits.keys()):
            self.remove_circuit(circuit_id, 'unload', remove_now=True, destroy=True)
        for circuit_id in list(self.relay_from_to.keys()):
            self.remove_relay(circuit_id, 'unload', remove_now=True, destroy=True, both_sides=False)
        for circuit_id in list(self.exit_sockets.keys()):
            self.remove_exit_socket(circuit_id, 'unload', remove_now=True, destroy=True)

        self.request_cache.shutdown()

        super(TunnelCommunity, self).unload()

    def get_session_keys(self, keys, direction):
        # increment salt_explicit
        keys[direction + 4] += 1
        return keys[direction], keys[direction + 2], keys[direction + 4]

    def _generate_circuit_id(self):
        circuit_id = random.getrandbits(32)

        # Prevent collisions.
        while circuit_id in self.circuits:
            circuit_id = random.getrandbits(32)

        return circuit_id

    def do_circuits(self):
        for circuit_length, num_circuits in self.circuits_needed.items():
            num_to_build = max(0, num_circuits - len(self.find_circuits(state=None, hops=circuit_length)))
            self.logger.info("Want %d data circuits of length %d", num_to_build, circuit_length)
            for _ in range(num_to_build):
                if not self.create_circuit(circuit_length):
                    self.logger.info("circuit creation of %d circuits failed, no need to continue", num_to_build)
                    break
        self.do_remove()

    def build_tunnels(self, hops):
        if hops > 0:
            self.circuits_needed[hops] = max(self.settings.max_circuits,
                                             min(self.settings.max_circuits, self.circuits_needed.get(hops, 0) + 1))
            self.do_circuits()

    def tunnels_ready(self, hops):
        if hops > 0 and self.circuits_needed.get(hops, 0):
            return len(self.find_circuits(hops=hops)) / float(self.circuits_needed[hops])
        return 1.0

    def do_remove(self):
        # Remove circuits that are inactive / are too old / have transferred too many bytes.
        for circuit_id, circuit in self.circuits.items():
            if circuit.state == CIRCUIT_STATE_READY and \
               circuit.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_circuit(circuit_id, 'no activity')
            elif circuit.creation_time < time.time() - self.get_max_time(circuit_id):
                self.remove_circuit(circuit_id, 'too old')
            elif circuit.bytes_up + circuit.bytes_down > self.settings.max_traffic:
                self.remove_circuit(circuit_id, 'traffic limit exceeded')

        # Remove relays that are inactive / have transferred too many bytes.
        for circuit_id, relay in self.relay_from_to.items():
            if relay.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_relay(circuit_id, 'no activity', both_sides=False)
            elif relay.bytes_up + relay.bytes_down > self.settings.max_traffic:
                self.remove_relay(circuit_id, 'traffic limit exceeded', both_sides=False)

        # Remove exit sockets that are too old / have transferred too many bytes.
        for circuit_id, exit_socket in self.exit_sockets.items():
            if exit_socket.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_exit_socket(circuit_id, 'no activity')
            elif exit_socket.creation_time < time.time() - self.get_max_time(circuit_id):
                self.remove_exit_socket(circuit_id, 'too old')
            elif exit_socket.bytes_up + exit_socket.bytes_down > self.settings.max_traffic:
                self.remove_exit_socket(circuit_id, 'traffic limit exceeded')

        # Remove candidates that are not returned as verified peers
        current_peers = set(self.network.get_peers_for_service(self.master_peer.mid))
        for peer in list(self.candidates):
            if peer not in current_peers:
                self.candidates.pop(peer)
                self.logger.info("Removed candidate from candidates dictionary")

    def get_candidates(self, flag):
        return [peer for peer, flags in self.candidates.items()
                if flags & flag and self.crypto.is_key_compatible(peer.public_key)]

    def get_max_time(self, circuit_id):
        return self.settings.max_time

    def find_circuits(self, ctype=CIRCUIT_TYPE_DATA, state=CIRCUIT_STATE_READY, hops=None):
        return [c for c in self.circuits.values()
                if (state is None or c.state == state)
                and (ctype is None or c.ctype == ctype)
                and (hops is None or hops == c.goal_hops)]

    def select_circuit(self, destination, hops):
        circuits = sorted(self.find_circuits(hops=hops), key=lambda c: c.circuit_id)
        if not circuits:
            return None

        self.select_index = (self.select_index + 1) % len(circuits)
        return circuits[self.select_index]

    def create_circuit(self, goal_hops, ctype=CIRCUIT_TYPE_DATA, callback=None, required_exit=None, info_hash=None):

        self.logger.info("Creating a new circuit of length %d (type: %s)", goal_hops, ctype)
        exit_candidates = self.get_candidates(PEER_FLAG_EXIT_ANY)
        if ctype == CIRCUIT_TYPE_IPV8:
            exit_candidates = self.get_candidates(PEER_FLAG_EXIT_IPV8) or exit_candidates
        relay_candidates = self.get_candidates(PEER_FLAG_RELAY)

        # Determine the last hop
        if not required_exit:
            if ctype in [CIRCUIT_TYPE_DATA, CIRCUIT_TYPE_IPV8, CIRCUIT_TYPE_IP_SEEDER]:
                required_exit = random.choice(exit_candidates) if exit_candidates else None
                # For introduction points we prefer exit nodes, but perhaps a relay peer would also suffice..
                if not required_exit and relay_candidates and ctype == CIRCUIT_TYPE_IP_SEEDER:
                    required_exit = random.choice(relay_candidates)
            else:
                # For exit nodes that don't exit actual data, we prefer relay candidates,
                # but we also consider exit candidates.
                if relay_candidates:
                    required_exit = random.choice(relay_candidates)
                elif exit_candidates:
                    required_exit = random.choice(exit_candidates)

        if not required_exit:
            self.logger.info("Could not create circuit, no available exit-nodes")
            return False

        # Determine the first hop
        if goal_hops == 1 and required_exit:
            # If the number of hops is 1, it should immediately be the required_exit hop.
            self.logger.info("First hop is required exit")
            possible_first_hops = [required_exit]
        else:
            self.logger.info("Look for a first hop that is not an exit node and is not used before")
            first_hops = set([c.peer.address for c in self.circuits.values() if c.peer])
            possible_first_hops = [c for c in relay_candidates if c.address not in first_hops
                                   and c.address != required_exit.address]

        if not possible_first_hops:
            self.logger.info("Could not create circuit, no first hop available")
            return False

        # Finally, construct the Circuit object and send the CREATE message
        circuit_id = self._generate_circuit_id()
        self.circuits[circuit_id] = circuit = Circuit(circuit_id, goal_hops, ctype, callback, required_exit, info_hash)
        self.send_initial_create(circuit, possible_first_hops,
                                 self.settings.circuit_timeout // self.settings.next_hop_timeout)

        return circuit_id

    def send_initial_create(self, circuit, candidate_list, max_tries):
        if self.request_cache.has(u"retry", circuit.circuit_id):
            self.request_cache.pop(u"retry", circuit.circuit_id)

        first_hop = random.choice(candidate_list)
        alt_first_hops = [c for c in candidate_list if c != first_hop]

        circuit.unverified_hop = Hop(first_hop.public_key)
        circuit.unverified_hop.address = first_hop.address
        circuit.unverified_hop.dh_secret, circuit.unverified_hop.dh_first_part = self.crypto.generate_diffie_secret()

        self.logger.info("Adding first hop %s:%d to circuit %d", *(first_hop.address + (circuit.circuit_id,)))

        self.request_cache.add(RetryRequestCache(self, circuit, alt_first_hops, max_tries - 1,
                                                 self.send_initial_create, self.settings.next_hop_timeout))

        self.increase_bytes_sent(circuit, self.send_cell([first_hop],
                                                         u"create",
                                                         CreatePayload(circuit.circuit_id,
                                                                       self.my_peer.public_key.key_to_bin(),
                                                                       circuit.unverified_hop.dh_first_part)))

    def remove_circuit(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        """
        Remove a circuit and return a deferred that fires when all data associated with the circuit is destroyed.
        Optionally send a destroy message.
        """
        circuit_to_remove = self.circuits.get(circuit_id, None)
        if circuit_to_remove is None:
            self.logger.warning('Cannot remove unknown circuit %d', circuit_id)
            return

        remove_deferred = Deferred()
        self.logger.info("Removing %s circuit %d %s", circuit_to_remove.ctype, circuit_id, additional_info)

        if destroy:
            self.destroy_circuit(circuit_to_remove)

        def remove_circuit_info():
            circuit = self.circuits.pop(circuit_id, None)
            if circuit:
                self.logger.info("Removed circuit %d " + additional_info, circuit_id)

            # Clean up the directions dictionary
            self.directions.pop(circuit_id, None)

            remove_deferred.callback(circuit_id)

        circuit_to_remove.close()

        if self.settings.remove_tunnel_delay == 0 or remove_now:
            remove_circuit_info()
        elif not self.is_pending_task_active("remove_circuit_%s" % circuit_id):
            self.register_task("remove_circuit_%s" % circuit_id,
                               reactor.callLater(self.settings.remove_tunnel_delay, remove_circuit_info))

        return remove_deferred

    def remove_relay(self, circuit_id, additional_info='', remove_now=False, destroy=False,
                     got_destroy_from=None, both_sides=True):
        """
        Remove a relay and all information associated with the relay. Return the relays that have been removed.
        """
        to_remove = [circuit_id]
        if both_sides:
            # Find other side of relay
            for k, v in self.relay_from_to.items():
                if circuit_id == v.circuit_id:
                    to_remove.append(k)

        # Send destroy
        if destroy:
            self.destroy_relay(to_remove, got_destroy_from=got_destroy_from)

        removed_relays = []
        for cid in to_remove:
            def remove_relay_info(cid_to_remove):
                # Remove the relay
                self.logger.info("Removing relay %d %s", cid_to_remove, additional_info)

                relay = self.relay_from_to.pop(cid_to_remove, None)
                if relay:
                    removed_relays.append(relay)

                # Remove old session key
                self.relay_session_keys.pop(cid_to_remove, None)

                # Clean directions dictionary
                self.directions.pop(cid_to_remove, None)

            if self.settings.remove_tunnel_delay == 0 or remove_now:
                remove_relay_info(cid)
            elif not self.is_pending_task_active("remove_relay_%s" % cid):
                self.register_task("remove_relay_%s" % cid,
                                   reactor.callLater(self.settings.remove_tunnel_delay,
                                                     lambda cid_copy=cid: remove_relay_info(cid_copy)))

        return removed_relays

    def remove_exit_socket(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        """
        Remove an exit socket. Send a destroy message if necessary.
        """
        exit_socket_to_destroy = self.exit_sockets.get(circuit_id, None)
        if exit_socket_to_destroy and destroy:
            self.destroy_exit_socket(exit_socket_to_destroy)

        def remove_exit_socket_info():
            exit_socket = self.exit_sockets.pop(circuit_id, None)
            if exit_socket:
                # Close socket
                if exit_socket.enabled:
                    self.logger.info("Removing exit socket %d %s", circuit_id, additional_info)

                    def on_exit_socket_closed(_):
                        # Remove old session key
                        self.relay_session_keys.pop(circuit_id, None)

                    addCallback(exit_socket.close(), on_exit_socket_closed)

        if self.settings.remove_tunnel_delay == 0 or remove_now:
            remove_exit_socket_info()
        elif not self.is_pending_task_active("remove_exit_socket_%s" % circuit_id):
            self.register_task("remove_exit_socket_%s" % circuit_id,
                               reactor.callLater(self.settings.remove_tunnel_delay, remove_exit_socket_info))

    def destroy_circuit(self, circuit, reason=0):
        sock_addr = circuit.peer.address
        self.send_destroy(sock_addr, circuit.circuit_id, reason)
        self.logger.info("destroy_circuit %s %s", circuit.circuit_id, sock_addr)

    def destroy_relay(self, circuit_ids, reason=0, got_destroy_from=None):
        relays = {cid_from: (self.relay_from_to[cid_from].circuit_id,
                             self.relay_from_to[cid_from].peer.address) for cid_from in circuit_ids
                  if cid_from in self.relay_from_to}

        if got_destroy_from and got_destroy_from not in relays.values():
            self.logger.error("%s not allowed send destroy for circuit %s", *reversed(got_destroy_from))
            return

        for cid_from, (cid_to, sock_addr) in relays.items():
            self.logger.info("Found relay %s -> %s (%s)", cid_from, cid_to, sock_addr)
            if (cid_to, sock_addr) != got_destroy_from:
                self.send_destroy(sock_addr, cid_to, reason)
                self.logger.info("Fw destroy to %s %s", cid_to, sock_addr)

    def destroy_exit_socket(self, exit_socket, reason=0):
        sock_addr = exit_socket.peer.address
        self.send_destroy(sock_addr, exit_socket.circuit_id, reason)
        self.logger.info("Destroy_exit_socket %s %s", exit_socket.circuit_id, sock_addr)

    def is_relay(self, circuit_id):
        return circuit_id > 0 and circuit_id in self.relay_from_to

    def is_circuit(self, circuit_id):
        return circuit_id > 0 and circuit_id in self.circuits

    def is_exit(self, circuit_id):
        return circuit_id > 0 and circuit_id in self.exit_sockets

    def send_cell(self, candidates, message_type, payload, circuit_id=None):
        message_id, _ = message_to_payload[message_type]
        circuit_id = circuit_id or payload.circuit_id
        if isinstance(payload, DataPayload):
            message = payload.to_bin()
        else:
            message = self.serializer.pack_multiple(payload.to_pack_list()[1:])[0]
        cell = CellPayload(circuit_id, pack('!B', message_id) + message, message_id in NO_CRYPTO_PACKETS)
        try:
            cell.encrypt(self.crypto, self.circuits.get(circuit_id), self.relay_session_keys.get(circuit_id))
        except CryptoException as e:
            self.logger.warning(str(e))
            return
        packet = cell.to_bin(self._prefix)
        return self.send_packet(candidates, packet)

    def send_data(self, candidates, circuit_id, dest_address, source_address, data):
        payload = DataPayload(circuit_id, dest_address, source_address, data)
        return self.send_cell(candidates, u"data", payload, circuit_id)

    def send_packet(self, candidates, packet):
        for candidate in candidates:
            address = candidate if isinstance(candidate, tuple) else candidate.address
            self.endpoint.send(address, packet)
        return len(packet)

    def send_destroy(self, candidate, circuit_id, reason):
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
        payload = DestroyPayload(circuit_id, reason).to_pack_list()
        packet = self._ez_pack(self._prefix, 10, [auth, payload])
        self.send_packet([candidate], packet)

    def relay_cell(self, cell):
        next_relay = self.relay_from_to[cell.circuit_id]
        try:
            if next_relay.rendezvous_relay:
                cell.decrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
                cell.encrypt(self.crypto, relay_session_keys=self.relay_session_keys[next_relay.circuit_id])
            else:
                direction = self.directions[cell.circuit_id]
                if direction == ORIGINATOR:
                    cell.encrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
                elif direction == EXIT_NODE:
                    cell.decrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
        except CryptoException as e:
            self.logger.warning(str(e))
            return

        cell.circuit_id = next_relay.circuit_id
        packet = cell.to_bin(self._prefix)
        self.increase_bytes_sent(next_relay, self.send_packet([next_relay.peer], packet))

    def _ours_on_created_extended(self, circuit, payload):
        hop = circuit.unverified_hop

        try:
            shared_secret = self.crypto.verify_and_generate_shared_secret(hop.dh_secret, payload.key,
                                                                          payload.auth, hop.public_key.key.pk)
            hop.session_keys = self.crypto.generate_session_keys(shared_secret)

        except CryptoException:
            self.remove_circuit(circuit.circuit_id, "error while verifying shared secret, bailing out.")
            return

        circuit.add_hop(hop)
        circuit.unverified_hop = None

        if circuit.state == CIRCUIT_STATE_EXTENDING:
            candidate_list_enc = payload.candidate_list_enc
            _, candidate_list = decode(self.crypto.decrypt_str(candidate_list_enc,
                                                               hop.session_keys[EXIT_NODE],
                                                               hop.session_keys[EXIT_NODE_SALT]))
            cache = self.request_cache.get(u"retry", payload.circuit_id)
            self.send_extend(circuit, candidate_list, cache.max_tries if cache else 1)

        elif circuit.state == CIRCUIT_STATE_READY:
            self.request_cache.pop(u"retry", payload.circuit_id)

            # Execute callback
            if circuit.callback:
                circuit.callback(circuit)
                circuit.callback = None

    def send_extend(self, circuit, candidate_list, max_tries):
        if self.request_cache.has(u"retry", circuit.circuit_id):
            self.request_cache.pop(u"retry", circuit.circuit_id)

        ignore_candidates = [self.crypto.key_to_bin(chop.public_key) for chop in circuit.hops] + \
                            [self.my_peer.public_key]
        if circuit.required_exit:
            ignore_candidates.append(circuit.required_exit.public_key.key_to_bin())

        become_exit = circuit.goal_hops - 1 == len(circuit.hops)
        if become_exit and circuit.required_exit:
            # Set the required exit according to the circuit setting (e.g. for linking e2e circuits)
            extend_hop_public_bin = circuit.required_exit.public_key.key_to_bin()
            extend_hop_addr = circuit.required_exit.address

        else:
            # The next candidate is chosen from the returned list of possible candidates
            for ignore_candidate in ignore_candidates:
                if ignore_candidate in candidate_list:
                    candidate_list.remove(ignore_candidate)

            for i in range(len(candidate_list) - 1, -1, -1):
                public_key = self.crypto.key_from_public_bin(candidate_list[i])
                if not self.crypto.is_key_compatible(public_key):
                    candidate_list.pop(i)

            extend_hop_public_bin = next(iter(candidate_list), None)
            extend_hop_addr = None

        if extend_hop_public_bin:
            extend_hop_public_key = self.crypto.key_from_public_bin(extend_hop_public_bin)
            circuit.unverified_hop = Hop(extend_hop_public_key)
            circuit.unverified_hop.dh_secret, circuit.unverified_hop.dh_first_part = \
                self.crypto.generate_diffie_secret()

            self.logger.info("Extending circuit %d with %s", circuit.circuit_id, hexlify(extend_hop_public_bin))

            # Only retry if we are allowed to use another node
            if not become_exit or not circuit.required_exit:
                alt_candidates = [c for c in candidate_list if c != extend_hop_public_bin]
            else:
                alt_candidates = []
            self.request_cache.add(RetryRequestCache(self, circuit, alt_candidates, max_tries - 1,
                                                     self.send_extend, self.settings.next_hop_timeout))

            self.increase_bytes_sent(circuit, self.send_cell([circuit.peer],
                                                             u"extend",
                                                             ExtendPayload(circuit.circuit_id,
                                                                           circuit.unverified_hop.node_public_key,
                                                                           extend_hop_addr,
                                                                           circuit.unverified_hop.dh_first_part)))

        else:
            self.remove_circuit(circuit.circuit_id, "no candidates to extend, bailing out.")

    def extract_peer_flags(self, extra_bytes):
        if not extra_bytes:
            return 0

        payload = self.serializer.unpack_to_serializables([ExtraIntroductionPayload], extra_bytes)[0]
        return payload.flags

    def introduction_request_callback(self, peer, dist, payload):
        self.candidates[peer] = self.extract_peer_flags(payload.extra_bytes)

    def introduction_response_callback(self, peer, dist, payload):
        self.candidates[peer] = self.extract_peer_flags(payload.extra_bytes)

    def create_introduction_request(self, socket_address, extra_bytes=b''):
        extra_payload = ExtraIntroductionPayload(self.settings.peer_flags)
        extra_bytes = self.serializer.pack_multiple(extra_payload.to_pack_list())[0]
        return super(TunnelCommunity, self).create_introduction_request(socket_address, extra_bytes)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier,
                                     introduction=None, extra_bytes=b''):
        extra_payload = ExtraIntroductionPayload(self.settings.peer_flags)
        extra_bytes = self.serializer.pack_multiple(extra_payload.to_pack_list())[0]
        return super(TunnelCommunity, self).create_introduction_response(lan_socket_address, socket_address,
                                                                         identifier, introduction, extra_bytes)

    def on_cell(self, source_address, data):
        cell = CellPayload.from_bin(data)
        circuit_id = cell.circuit_id

        if self.is_relay(circuit_id):
            next_relay = self.relay_from_to[circuit_id]
            this_relay = self.relay_from_to.get(next_relay.circuit_id, None)
            if this_relay:
                this_relay.beat_heart()
                self.increase_bytes_received(this_relay, len(data))
            self.logger.debug("Relaying cell from circuit %d to %d", circuit_id, next_relay.circuit_id)
            self.relay_cell(cell)
            return

        circuit = self.circuits.get(circuit_id, None)
        try:
            cell.decrypt(self.crypto, circuit=circuit, relay_session_keys=self.relay_session_keys.get(circuit_id))
        except CryptoException:
            if circuit:
                self.send_destroy(circuit.peer, circuit_id, 0)
            return
        self.logger.debug("Got cell(%s) from circuit %d (sender %s, receiver %s)",
                          cell.message[0], circuit_id, source_address, self.my_peer)
        self.on_packet_from_circuit(source_address, cell.unwrap(self._prefix), circuit_id)

        if circuit:
            circuit.beat_heart()
            self.increase_bytes_received(circuit, len(data))

    def on_packet_from_circuit(self, source_address, data, circuit_id):
        if self._prefix != data[:22]:
            return
        msg_id = chr(ord(data[22:23]))
        if msg_id in self.decode_map_private:
            try:
                self.decode_map_private[msg_id](source_address, data, circuit_id)
            except:
                self.logger.error("Exception occurred while handling packet!\n"
                                  + ''.join(format_exception(*sys.exc_info())))

    def should_join_circuit(self, create_payload, previous_node_address):
        """
        Check whether we should join a circuit.
        Returns a deferred that fires with a boolean.
        """
        if self.settings.max_joined_circuits <= len(self.relay_from_to) + len(self.exit_sockets):
            self.logger.warning("Too many relays (%d)", (len(self.relay_from_to) + len(self.exit_sockets)))
            return succeed(False)

        return succeed(True)

    def join_circuit(self, create_payload, previous_node_address):
        """
        Actively join a circuit and send a created message back
        """
        circuit_id = create_payload.circuit_id

        self.directions[circuit_id] = EXIT_NODE
        self.logger.info('We joined circuit %d with neighbour %s', circuit_id, previous_node_address)

        shared_secret, key, auth = self.crypto.generate_diffie_shared_secret(create_payload.key)
        self.relay_session_keys[circuit_id] = self.crypto.generate_session_keys(shared_secret)

        peers_list = [peer for peer in self.get_candidates(PEER_FLAG_RELAY)
                      if peer not in self.get_candidates(PEER_FLAG_EXIT_ANY)][:4]
        peers_keys = {c.public_key.key_to_bin(): c for c in peers_list}

        peer = Peer(create_payload.node_public_key, previous_node_address)
        self.request_cache.add(CreatedRequestCache(self, circuit_id, peer, peers_keys, self.settings.unstable_timeout))
        self.exit_sockets[circuit_id] = TunnelExitSocket(circuit_id, peer, self)

        candidate_list_enc = self.crypto.encrypt_str(encode(list(peers_keys.keys())),
                                                     *self.get_session_keys(self.relay_session_keys[circuit_id],
                                                                            EXIT_NODE))
        self.send_cell([Peer(create_payload.node_public_key, previous_node_address)], u"created",
                       CreatedPayload(circuit_id, key, auth, candidate_list_enc))

    @tc_lazy_wrapper_unsigned(CreatePayload)
    def on_create(self, source_address, payload, _):
        if not self.settings.peer_flags:
            self.logger.warning("Ignoring create for circuit %d", payload.circuit_id)
            return
        if self.request_cache.has(u"created", payload.circuit_id):
            self.logger.warning("Already have a request for circuit %d", payload.circuit_id)
            return

        def determined_to_join(result):
            if result:
                self.join_circuit(payload, source_address)
            else:
                self.logger.warning("We're not joining circuit with ID %s", payload.circuit_id)

        addCallback(self.should_join_circuit(payload, source_address), determined_to_join)

    @tc_lazy_wrapper_unsigned(CreatedPayload)
    def on_created(self, source_address, payload, _):
        circuit_id = payload.circuit_id
        self.directions[circuit_id] = ORIGINATOR

        if self.request_cache.has(u"create", payload.circuit_id):
            request = self.request_cache.pop(u"create", circuit_id)

            self.logger.info("Got CREATED message forward as EXTENDED to origin.")

            self.relay_from_to[request.to_circuit_id] = relay = RelayRoute(request.from_circuit_id, request.peer)
            self.relay_from_to[request.from_circuit_id] = RelayRoute(request.to_circuit_id, request.to_peer)
            self.relay_session_keys[request.to_circuit_id] = self.relay_session_keys[request.from_circuit_id]

            self.directions[request.from_circuit_id] = EXIT_NODE
            self.remove_exit_socket(request.from_circuit_id)

            self.send_cell([relay.peer], u"extended", ExtendedPayload(relay.circuit_id,
                                                                      payload.key,
                                                                      payload.auth,
                                                                      payload.candidate_list_enc))
        elif self.request_cache.has(u"retry", payload.circuit_id):
            circuit = self.circuits[circuit_id]
            self._ours_on_created_extended(circuit, payload)
        else:
            self.logger.warning("Received unexpected created for circuit %d", payload.circuit_id)

    @inlineCallbacks
    @tc_lazy_wrapper_unsigned(ExtendPayload)
    def on_extend(self, source_address, payload, _):
        if not self.settings.peer_flags & PEER_FLAG_RELAY:
            self.logger.warning("Ignoring create for circuit %d", payload.circuit_id)
            returnValue(None)
        if not self.request_cache.has(u"created", payload.circuit_id):
            self.logger.warning("Received unexpected extend for circuit %d", payload.circuit_id)
            returnValue(None)

        circuit_id = payload.circuit_id
        # Leave the RequestCache in case the circuit owner wants to reuse the tunnel for a different next-hop
        request = self.request_cache.get(u"created", circuit_id)
        if not (payload.node_addr or payload.node_public_key in request.candidates):
            self.logger.warning("Node public key not in request candidates and no ip specified")
            returnValue(None)

        if payload.node_public_key in request.candidates:
            extend_candidate = request.candidates[payload.node_public_key]
        else:
            extend_candidate = self.network.get_verified_by_public_key_bin(payload.node_public_key)
            if not extend_candidate:
                extend_candidate = Peer(payload.node_public_key, payload.node_addr)

            # Ensure that we are able to contact this peer
            if extend_candidate.last_response + 57.5 < time.time():
                deferred = Deferred()
                self.dht_peer_lookup(extend_candidate.mid, lambda _: deferred.callback(None))
                yield deferred

        self.logger.info("On_extend send CREATE for circuit (%s, %d) to %s:%d", source_address,
                         circuit_id, *extend_candidate.address)

        to_circuit_id = self._generate_circuit_id()

        if circuit_id in self.circuits:
            candidate = self.circuits[circuit_id].peer
        elif circuit_id in self.exit_sockets:
            candidate = self.exit_sockets[circuit_id].peer
        elif circuit_id in self.relay_from_to:
            candidate = self.relay_from_to[circuit_id].peer
        else:
            self.logger.error("Got extend for unknown source circuit_id")
            returnValue(None)

        self.logger.info("Extending circuit, got candidate with IP %s:%d from cache", *extend_candidate.address)

        self.request_cache.add(CreateRequestCache(self, to_circuit_id, circuit_id,
                                                  candidate, extend_candidate))

        self.send_cell([extend_candidate], u"create",
                       CreatePayload(to_circuit_id, self.my_peer.public_key.key_to_bin(), payload.key))

    @tc_lazy_wrapper_unsigned(ExtendedPayload)
    def on_extended(self, source_address, payload, _):
        if not self.request_cache.has(u"retry", payload.circuit_id):
            self.logger.warning("Received unexpected extended for circuit %s", payload.circuit_id)
            return

        circuit_id = payload.circuit_id
        circuit = self.circuits[circuit_id]
        self._ours_on_created_extended(circuit, payload)

    def on_raw_data(self, circuit, origin, data):
        """
        Handle data, coming from a specific circuit and origin.
        This method is usually implemented in subclasses of this community.
        """
        pass

    def on_data(self, sock_addr, data, _):
        payload = DataPayload.from_bin(data)

        # If its our circuit, the messenger is the candidate assigned to that circuit and the DATA's destination
        # is set to the zero-address then the packet is from the outside world and addressed to us from.
        circuit_id = payload.circuit_id
        destination = payload.dest_address
        origin = payload.org_address
        data = payload.data

        self.logger.debug("Got data (%d) from %s", circuit_id, sock_addr)

        circuit = self.circuits.get(circuit_id, None)
        if circuit and origin and sock_addr == circuit.peer.address:
            circuit.beat_heart()

            if DataChecker.could_be_ipv8(data):
                if self._prefix == data[:22]:
                    self.logger.debug("Incoming packet meant for us")
                    self.on_packet_from_circuit(origin, data, circuit_id)
                    return

                if isinstance(self.endpoint, TunnelEndpoint):
                    self.logger.debug("Incoming packet meant for other community")
                    self.endpoint.notify_listeners((origin, data), from_tunnel=True)
                else:
                    self.logger.debug("Incoming packet meant for other community, dropping")
            else:
                # We probably received raw data, handle it
                self.on_raw_data(circuit, origin, data)

        # It is not our circuit so we got it from a relay, we need to EXIT it!
        else:
            self.logger.debug("Data for circuit %d exiting tunnel (%s)", circuit_id, destination)
            if destination != ('0.0.0.0', 0):
                self.exit_data(circuit_id, sock_addr, destination, data)
            else:
                self.logger.warning("Cannot exit data, destination is 0.0.0.0:0")

    @tc_lazy_wrapper_unsigned(PingPayload)
    def on_ping(self, source_address, payload, _):
        if not (payload.circuit_id in self.circuits
                or payload.circuit_id in self.exit_sockets
                or payload.circuit_id in self.relay_from_to):
            return

        exit_socket = self.exit_sockets.get(payload.circuit_id)
        if exit_socket:
            exit_socket.beat_heart()

        self.send_cell([source_address], u"pong", PongPayload(payload.circuit_id, payload.identifier))
        self.logger.debug("Got ping from %s", source_address)

    @tc_lazy_wrapper_unsigned(PongPayload)
    def on_pong(self, source_address, payload, _):
        if not self.request_cache.has(u"ping", payload.identifier):
            self.logger.warning("Invalid ping circuit_id")
            return

        self.request_cache.pop(u"ping", payload.identifier)
        self.logger.debug("Got pong from %s", source_address)

    def do_ping(self, exclude=None):
        # Ping circuits. Pings are only sent to the first hop, subsequent hops will relay the ping.
        exclude = [] if exclude is None else exclude
        for circuit in self.circuits.values():
            if circuit.state in [CIRCUIT_STATE_READY, CIRCUIT_STATE_EXTENDING] \
                    and circuit.circuit_id not in exclude \
                    and circuit.hops:
                cache = self.request_cache.add(PingRequestCache(self, circuit))
                self.increase_bytes_sent(circuit, self.send_cell([circuit.peer], u"ping",
                                                                 PingPayload(circuit.circuit_id, cache.number)))

    @lazy_wrapper(DestroyPayload)
    def on_destroy(self, peer, payload):
        source_address = peer.address
        circuit_id = payload.circuit_id
        self.logger.info("Got destroy from %s for circuit %s", source_address, circuit_id)

        if circuit_id in self.relay_from_to:
            self.remove_relay(circuit_id, "Got destroy", destroy=True, got_destroy_from=(circuit_id, source_address))

        elif circuit_id in self.exit_sockets and source_address == self.exit_sockets[circuit_id].peer.address:
            self.logger.info("Got an exit socket %s %s", circuit_id, source_address)
            self.remove_exit_socket(circuit_id, "Got destroy")

        elif circuit_id in self.circuits and source_address == self.circuits[circuit_id].peer.address:
            self.logger.info("Got a circuit %s %s", circuit_id, source_address)
            self.remove_circuit(circuit_id, "Got destroy")

        else:
            self.logger.warning("Invalid or unauthorized destroy")

    def exit_data(self, circuit_id, sock_addr, destination, data):
        is_ipv8 = DataChecker.could_be_ipv8(data)
        is_ipv8_tunnel = is_ipv8 and self._prefix == data[:22]

        if not (is_ipv8 or self.settings.peer_flags & PEER_FLAG_EXIT_ANY):
            self.logger.error("Dropping data packets, refusing to be an exit node for data")

        elif not (is_ipv8_tunnel
                  or self.settings.peer_flags & PEER_FLAG_EXIT_IPV8
                  or self.settings.peer_flags & PEER_FLAG_EXIT_ANY):
            self.logger.error("Dropping data packets, refusing to be an exit node for ipv8")

        elif circuit_id in self.exit_sockets:
            if not self.exit_sockets[circuit_id].enabled:
                # Check that we got the data from the correct IP.
                if sock_addr[0] == self.exit_sockets[circuit_id].peer.address[0]:
                    self.exit_sockets[circuit_id].enable()
                else:
                    self.logger.error("Dropping outbound relayed packet: IP's are %s != %s",
                                      str(sock_addr), str(self.exit_sockets[circuit_id].peer.address))
            try:
                self.exit_sockets[circuit_id].sendto(data, destination)
            except:
                self.logger.warning("Dropping data packets while EXITing")
        else:
            self.logger.error("Dropping data packets with unknown circuit_id")

    def increase_bytes_sent(self, obj, num_bytes):
        obj.bytes_up += num_bytes

    def increase_bytes_received(self, obj, num_bytes):
        obj.bytes_down += num_bytes

    def dht_peer_lookup(self, mid, cb):
        if self.dht_provider:
            self.dht_provider.peer_lookup(mid, cb)
        else:
            self.logger.error("Need a DHT provider to connect to a peer using the DHT")



from __future__ import absolute_import

from binascii import unhexlify

from twisted.web import http

from .base_endpoint import BaseEndpoint
from ..attestation.trustchain.community import TrustChainCommunity


class TrustchainEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding TrustChain.
    """

    def __init__(self, session):
        super(TrustchainEndpoint, self).__init__()

        trustchain_overlays = [overlay for overlay in session.overlays if isinstance(overlay, TrustChainCommunity)]
        if trustchain_overlays:
            self.putChild("recent", TrustchainRecentEndpoint(trustchain_overlays[0]))
            self.putChild("blocks", TrustchainBlocksEndpoint(trustchain_overlays[0]))
            self.putChild("users", TrustchainUsersEndpoint(trustchain_overlays[0]))


class TrustchainRecentEndpoint(BaseEndpoint):

    def __init__(self, trustchain):
        super(TrustchainRecentEndpoint, self).__init__()
        self.trustchain = trustchain

    def render_GET(self, request):
        limit = 10
        offset = 0
        if request.args and 'limit' in request.args:
            limit = int(request.args['limit'][0])

        if request.args and 'offset' in request.args:
            offset = int(request.args['offset'][0])

        return self.twisted_dumps({
            "blocks": [dict(block) for block in
                       self.trustchain.persistence.get_recent_blocks(limit=limit, offset=offset)]
        })


class TrustchainBlocksEndpoint(BaseEndpoint):

    def __init__(self, trustchain):
        super(TrustchainBlocksEndpoint, self).__init__()
        self.trustchain = trustchain

    def getChild(self, path, request):
        return TrustchainSpecificBlockEndpoint(self.trustchain, path)


class TrustchainSpecificBlockEndpoint(BaseEndpoint):

    def __init__(self, trustchain, block_hash):
        super(TrustchainSpecificBlockEndpoint, self).__init__()
        self.trustchain = trustchain
        try:
            self.block_hash = unhexlify(block_hash)
        except TypeError:
            self.block_hash = None

    def render_GET(self, request):
        if not self.block_hash:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "the block with the provided hash could not be found"})

        block = self.trustchain.persistence.get_block_with_hash(self.block_hash)
        if not block:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "the block with the provided hash could not be found"})

        block_dict = dict(block)

        # Fetch the linked block if available
        linked_block = self.trustchain.persistence.get_linked(block)
        if linked_block:
            block_dict["linked"] = dict(linked_block)

        return self.twisted_dumps({"block": block_dict})


class TrustchainUsersEndpoint(BaseEndpoint):

    def __init__(self, trustchain):
        super(TrustchainUsersEndpoint, self).__init__()
        self.trustchain = trustchain

    def getChild(self, path, request):
        return TrustchainSpecificUserEndpoint(self.trustchain, path)

    def render_GET(self, request):
        limit = 100
        if 'limit' in request.args:
            limit = int(request.args['limit'][0])

        users_info = self.trustchain.persistence.get_users(limit=limit)
        return self.twisted_dumps({"users": users_info})


class TrustchainSpecificUserEndpoint(BaseEndpoint):

    def __init__(self, trustchain, pub_key):
        super(TrustchainSpecificUserEndpoint, self).__init__()
        self.trustchain = trustchain
        self.pub_key = pub_key

        self.putChild("blocks", TrustchainSpecificUserBlocksEndpoint(self.trustchain, self.pub_key))


class TrustchainSpecificUserBlocksEndpoint(BaseEndpoint):

    def __init__(self, trustchain, pub_key):
        super(TrustchainSpecificUserBlocksEndpoint, self).__init__()
        self.trustchain = trustchain
        try:
            self.pub_key = unhexlify(pub_key)
        except TypeError:
            self.pub_key = None

    def render_GET(self, request):
        if not self.pub_key:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "the user with the provided public key could not be found"})

        limit = 100
        if 'limit' in request.args:
            limit = int(request.args['limit'][0])

        latest_blocks = self.trustchain.persistence.get_latest_blocks(self.pub_key, limit=limit)
        blocks_list = []
        for block in latest_blocks:
            block_dict = dict(block)
            linked_block = self.trustchain.persistence.get_linked(block)
            if linked_block:
                block_dict['linked'] = dict(linked_block)
            blocks_list.append(block_dict)

        return self.twisted_dumps({"blocks": blocks_list})



from __future__ import absolute_import

from collections import OrderedDict

from twisted.trial import unittest

from ....messaging.deprecated.sorting import sortable_sort


class TestSorting(unittest.TestCase):

    def test_illegal_object(self):
        self.assertRaises(RuntimeError, sortable_sort, self)

    def test_type_ordering(self):
        """
        Check if types form the primary sorting key.
        """
        out = sortable_sort(["1", None, [1, 2], True, 1])

        self.assertListEqual(out, [None, True, 1, "1", [1, 2]])

    def test_single_none(self):
        """
        Check if a None is returned as is.
        """
        self.assertEqual(sortable_sort(None), None)

    def test_single_bool_true(self):
        """
        Check if a True is returned as is.
        """
        self.assertEqual(sortable_sort(True), True)

    def test_single_bool_false(self):
        """
        Check if a False is returned as is.
        """
        self.assertEqual(sortable_sort(False), False)

    def test_single_float_zero(self):
        """
        Check if a 0.0 is returned as is.
        """
        self.assertEqual(sortable_sort(0.0), 0.0)

    def test_single_float_minus_one(self):
        """
        Check if a -1.0 is returned as is.
        """
        self.assertEqual(sortable_sort(-1.0), -1.0)

    def test_single_float_plus_one(self):
        """
        Check if a 1.0 is returned as is.
        """
        self.assertEqual(sortable_sort(1.0), 1.0)

    def test_single_float_large(self):
        """
        Check if a 99999.9999 is returned as is.
        """
        self.assertEqual(sortable_sort(99999.9999), 99999.9999)

    def test_single_int_zero(self):
        """
        Check if a 0 is returned as is.
        """
        self.assertEqual(sortable_sort(0), 0)

    def test_single_int_minus_one(self):
        """
        Check if a -1 is returned as is.
        """
        self.assertEqual(sortable_sort(-1), -1)

    def test_single_int_plus_one(self):
        """
        Check if a 1 is returned as is.
        """
        self.assertEqual(sortable_sort(1), 1)

    def test_single_float_int(self):
        """
        Check if a 999999999 is returned as is.
        """
        self.assertEqual(sortable_sort(999999999), 999999999)

    def test_single_bytes_empty(self):
        """
        Check if an empty byte string is returned as is.
        """
        self.assertEqual(sortable_sort(b""), b"")

    def test_single_bytes_one(self):
        """
        Check if a single character byte string is returned as is.
        """
        self.assertEqual(sortable_sort(b"\x01"), b"\x01")

    def test_single_bytes_many(self):
        """
        Check if an all different bytes string is returned as is.
        """
        self.assertEqual(sortable_sort(bytes(range(256))), bytes(range(256)))

    def test_single_str_empty(self):
        """
        Check if an empty string is returned as is.
        """
        self.assertEqual(sortable_sort(""), "")

    def test_single_str_one(self):
        """
        Check if a single character string is returned as is.
        """
        self.assertEqual(sortable_sort("\x01"), "\x01")

    def test_single_str_many(self):
        """
        Check if an all different character string is returned as is.
        """
        self.assertEqual(sortable_sort("".join([chr(c) for c in range(256)])), "".join([chr(c) for c in range(256)]))

    def test_single_tuple_empty(self):
        """
        Check if an empty tuple is returned as is.
        """
        self.assertTupleEqual(sortable_sort(tuple()), tuple())

    def test_single_tuple_one(self):
        """
        Check if a single entry tuple is returned as is.
        """
        self.assertTupleEqual(sortable_sort((b"\x01", )), (b"\x01", ))

    def test_single_tuple_many(self):
        """
        Check if a filled tuple is returned as is.
        """
        data = (5, 4, "z", "a", [4, 2])
        expected = (5, 4, "z", "a", [2, 4])

        self.assertTupleEqual(sortable_sort(data), expected)

    def test_single_set_empty(self):
        """
        Check if an empty set is returned as is.
        """
        self.assertSetEqual(sortable_sort(set()), set())
        self.assertListEqual(list(sortable_sort(set())), [])

    def test_single_set_one(self):
        """
        Check if a single entry set is returned as is.
        """
        self.assertSetEqual(sortable_sort(set([b"\x01"])), set([b"\x01"]))
        self.assertListEqual(list(sortable_sort({b"\x01"})), [b"\x01"])

    def test_single_set_many(self):
        """
        Check if a filled set is returned sorted.
        """
        data = {5, 4, "z", "a"}
        expected = [4, 5, "a", "z"]

        self.assertSetEqual(sortable_sort(data), set(expected))
        self.assertListEqual(list(sortable_sort(data)), expected)

    def test_single_list_empty(self):
        """
        Check if an empty list is returned as is.
        """
        self.assertListEqual(sortable_sort([]), [])

    def test_single_list_one(self):
        """
        Check if a single entry list is returned as is.
        """
        self.assertListEqual(sortable_sort([b"\x01"]), [b"\x01"])

    def test_single_list_many(self):
        """
        Check if a filled list is returned sorted.
        """
        data = [5, 4, "z", "a", [4, 2]]
        expected = [4, 5, "a", "z", [2, 4]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_single_dict_empty(self):
        """
        Check if an empty dict is returned as is.
        """
        self.assertDictEqual(sortable_sort({}), {})

    def test_single_dict_one(self):
        """
        Check if a single entry dict is returned as is.
        """
        self.assertDictEqual(sortable_sort({b"\x01": b"\x02"}), {b"\x01": b"\x02"})

    def test_single_dict_many(self):
        """
        Check if a filled dict is returned sorted.
        """
        data = {5: 4, "z": "a", 1: [4, 2]}
        expected = {1: [2, 4], 5: 4, "z": "a"}

        self.assertDictEqual(sortable_sort(data), expected)

    def test_nested_list_single(self):
        """
        Check if a list of lists is sorted.
        """
        data = [[5, 4], [2, 3], [1, 0]]
        expected = [[0, 1], [2, 3], [4, 5]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_equal(self):
        """
        Check if a list of equal lists is sorted.
        """
        data = [[0, 1], [1, 0]]
        expected = [[0, 1], [0, 1]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_partial(self):
        """
        Check if a list of lists with overlap is sorted.
        """
        data = [[1, 2], [0, 1, 2, 3], [2, 1, 0], [0, 1, 4, 3, 2]]
        expected = [[0, 1, 2], [0, 1, 2, 3], [0, 1, 2, 3, 4], [1, 2]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_double_mixed(self):
        """
        Check if a list of lists of half list is sorted.
        """
        data = [[5, 4], [[2, 2.5], 3], [1, 0]]
        expected = [[0, 1], [3, [2, 2.5]], [4, 5]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_double_pure(self):
        """
        Check if a list of lists of list is sorted.
        """
        data = [[5, 4], [[2, 2.5], [3, 3.5]], [1, 0]]
        expected = [[0, 1], [4, 5], [[2, 2.5], [3, 3.5]]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_bools(self):
        """
        Check if booleans are propertly sorted in a list.
        """
        data = [True, False, False, True, False]
        expected = [False, False, False, True, True]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_list_none(self):
        """
        Check if nones are propertly sorted in a list.
        """
        data = [None, None, [None]]
        expected = [None, None, [None]]

        self.assertListEqual(sortable_sort(data), expected)

    def test_nested_dict_single(self):
        """
        Check if a dict of lists is sorted.
        """
        data = {1: [5, 4], 3: [2, 3], 2: [1, 0]}
        expected = OrderedDict([(1, [4, 5]), (2, [0, 1]), (3, [2, 3])])

        self.assertDictEqual(sortable_sort(data), expected)

    def test_nested_dict_double(self):
        """
        Check if a dict of dicts is sorted.
        """
        data = {7: {2: 4, 4: 5}, 1: {4: 5, 2: 3}}
        expected = OrderedDict([(1, OrderedDict([(2, 3), (4, 5)])), (7, OrderedDict([(2, 4), (4, 5)]))])

        self.assertDictEqual(sortable_sort(data), expected)

    def test_payload_of_doom(self):
        """
        Check if an inconsistent sorted() list is sorted correctly.

        In other words:
            Given: set(data) == set(data2)

            Given: sorted(data) != sorted(data2)

            Then: sortable_sort(data) == sortable_sort(data2)
        """
        data = ['a', u'b', (1, ), 'c']
        data2 = ['a', u'b', 'c', (1, )]

        self.assertListEqual(sortable_sort(data), sortable_sort(data2))



from __future__ import absolute_import

import logging
from binascii import hexlify, unhexlify

from twisted.web import http

from .base_endpoint import BaseEndpoint
from ..dht.community import DHTCommunity


class NoBlockDHTEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT data, non-blocking.
    """

    def __init__(self, session):
        super(NoBlockDHTEndpoint, self).__init__()
        dht_overlays = [overlay for overlay in session.overlays if isinstance(overlay, DHTCommunity)]
        if dht_overlays:
            self.putChild(b"peers", NoBlockDHTPeersEndpoint(dht_overlays[0]))


class NoBlockDHTPeersEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT peers, non-blocking.
    """

    def __init__(self, dht):
        super(NoBlockDHTPeersEndpoint, self).__init__()
        self.dht = dht

    def getChild(self, path, request):
        return NoBlockSpecificDHTPeerEndpoint(self.dht, path)


class NoBlockSpecificDHTPeerEndpoint(BaseEndpoint):
    """
    This class handles requests for a specific DHT peer, non-blocking.
    """

    def __init__(self, dht, key):
        super(NoBlockSpecificDHTPeerEndpoint, self).__init__()
        self.mid = bytes(unhexlify(key))
        self.dht = dht

    def render_GET(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        def on_success(nodes):
            logging.error("DHT connected to %s", hexlify(self.mid))

        def on_failure(failure):
            logging.error("DHT Failed to connect to %s", hexlify(self.mid))

        self.dht.connect_peer(self.mid).addCallbacks(on_success, on_failure)

        return self.twisted_dumps({"success": True})



from __future__ import absolute_import

import threading

from twisted.internet.task import LoopingCall

from .comunities import TestIdentityCommunity, TestAttestationCommunity, overlay_initializer
from ....keyvault.crypto import ECCrypto
from ....messaging.interfaces.udp.endpoint import UDPEndpoint
from ....peer import Peer
from ....peerdiscovery.discovery import RandomWalk
from ....peerdiscovery.network import Network


class TestRestIPv8(object):

    def __init__(self, crypto_curve, port, interface, overlay_classes, memory_dbs=True):
        self.memory_dbs = memory_dbs
        self.endpoint = UDPEndpoint(port=port, ip=interface)
        self.endpoint.open()

        self.network = Network()

        my_peer = Peer(ECCrypto().generate_key(crypto_curve))
        self.keys = {'my_peer': my_peer}

        database_working_dir = u":memory:" if memory_dbs else u""
        self.overlays = []

        for overlay_class in overlay_classes:
            self.overlays.append(overlay_initializer(overlay_class, my_peer, self.endpoint, self.network,
                                                     working_directory=database_working_dir))

        self.strategies = [
            (RandomWalk(overlay), 20) for overlay in self.overlays
        ]

        self.overlay_lock = threading.RLock()
        self.state_machine_lc = LoopingCall(self.on_tick)
        self.state_machine_lc.start(0.5, False)

    def on_tick(self):
        if self.endpoint.is_open():
            with self.overlay_lock:
                for strategy, target_peers in self.strategies:
                    peer_count = len(strategy.overlay.get_peers())
                    if (target_peers == -1) or (peer_count < target_peers):
                        strategy.take_step()

    def unload(self):
        # Make sure the state machine is running before closing it
        if self.state_machine_lc.running:
            self.state_machine_lc.stop()

        if self.endpoint.is_open():
            self.endpoint.close()

        for overlay in self.overlays:
            # Close the DBs since simply unloading will usually not do
            if isinstance(overlay, TestAttestationCommunity):
                overlay.database.close()
            elif isinstance(overlay, TestIdentityCommunity):
                overlay.persistence.close()
            # Clear the cache manually
            overlay.request_cache.clear()
            overlay.unload()



class NetworkStat(object):
    """
    Represents an individual network statistic. This is used to compose the overall community statistics.
    Includes the following properties for the given statistic:
        - identifier: char
        - num_up: integer
        - num_down: integer
        - bytes_up: integer
        - bytes_down: integer
        - first_measured_up: float
        - first_measured_down: float
        - last_measured_up: float
        - last_measured_down: float
    """
    def __init__(self, identifier):
        self.identifier = identifier
        self.num_up = 0
        self.num_down = 0
        self.bytes_up = 0
        self.bytes_down = 0
        self.first_measured_up = 0
        self.first_measured_down = 0
        self.last_measured_up = 0
        self.last_measured_down = 0

    def add_sent_stat(self, timestamp, num_bytes):
        self.num_up += 1
        self.bytes_up += num_bytes
        self.last_measured_up = timestamp

        if not self.first_measured_up:
            self.first_measured_up = timestamp

    def add_received_stat(self, timestamp, num_bytes):
        self.num_down += 1
        self.bytes_down += num_bytes
        self.last_measured_down = timestamp

        if not self.first_measured_down:
            self.first_measured_down = timestamp

    def to_dict(self):
        return {
            "identifier": self.identifier,
            "num_up": self.num_up,
            "num_down": self.num_down,
            "bytes_up": self.bytes_up,
            "bytes_down": self.bytes_down,
            "first_measured_up": self.first_measured_up,
            "first_measured_down": self.first_measured_down,
            "last_measured_up": self.last_measured_up,
            "last_measured_down": self.last_measured_down
        }

    def __str__(self):
        return 'NetworkStat{num_up:%s, num_down:%s, bytes_up:%s, bytes_down:%s, ...}' % \
               (self.num_up, self.num_down, self.bytes_up, self.bytes_down)



from __future__ import absolute_import

from binascii import hexlify

from twisted.web import http

from .base_endpoint import BaseEndpoint
from ..messaging.interfaces.statistics_endpoint import StatisticsEndpoint


class OverlaysEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding the status of overlays.
    """

    def __init__(self, session):
        super(OverlaysEndpoint, self).__init__()
        self.session = session
        self.putChild("statistics", OverlayStatisticsEndpoint(session))

    def get_overlays(self):
        overlay_stats = []
        for overlay in self.session.overlays:
            peers = overlay.get_peers()
            statistics = self.session.endpoint.get_aggregate_statistics(overlay.get_prefix()) \
                if isinstance(self.session.endpoint, StatisticsEndpoint) else {}
            overlay_stats.append({
                "master_peer": hexlify(overlay.master_peer.public_key.key_to_bin()),
                "my_peer": hexlify(overlay.my_peer.public_key.key_to_bin()),
                "global_time": overlay.global_time,
                "peers": [str(peer) for peer in peers],
                "overlay_name": overlay.__class__.__name__,
                "statistics": statistics
            })
        return overlay_stats

    def render_GET(self, request):
        return self.twisted_dumps({"overlays": self.get_overlays()})


class OverlayStatisticsEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding the statistics of overlays.
    """

    def __init__(self, session):
        super(OverlayStatisticsEndpoint, self).__init__()
        self.session = session
        self.statistics_supported = isinstance(self.session.endpoint, StatisticsEndpoint)

    def get_statistics(self):
        overlay_stats = []
        for overlay in self.session.overlays:
            statistics = self.session.endpoint.get_statistics(overlay.get_prefix()) if self.statistics_supported else {}
            overlay_stats.append({
                overlay.__class__.__name__: self.statistics_by_name(statistics, overlay)
            })
        return overlay_stats

    def statistics_by_name(self, statistics, overlay):
        named_statistics = {}
        for message_id, network_stats in statistics.items():
            if overlay.decode_map.get(chr(message_id)):
                mapped_name = str(message_id) + ":" + overlay.decode_map[chr(message_id)].__name__
            else:
                mapped_name = str(message_id) + ":unknown"
            mapped_value = network_stats.to_dict()
            named_statistics[mapped_name] = mapped_value
        return named_statistics

    def render_GET(self, _):
        return self.twisted_dumps({"statistics": self.get_statistics()})

    def render_POST(self, request):
        """
        .. http:post:: /overlays/statistics

        A POST request to this endpoint will enable statistics on the given overlay.
        - enable: whether to enable or disable the statistics (True/False)
        - overlay_name: class name of the overlay
        - all: if set to True, update applies to all overlays

            **Example request**:

                .. sourcecode:: none

                    curl -X PUT http://localhost:8085/ipv8/overlays/statistics
                    --data "enable=True&overlay_name=overlay_name&all=True

            **Example response**:

                .. sourcecode:: javascript

                    {"success": True}
        """

        if not self.statistics_supported:
            request.setResponseCode(http.PRECONDITION_FAILED)
            return self.twisted_dumps({"success": False, "error": "StatisticsEndpoint is not enabled."})

        all_overlays = False
        overlay_name = None

        if 'enable' not in request.args or not request.args['enable']:
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"success": False, "error": "Parameter 'enable' is required"})
        else:
            enable = request.args['enable'][0] == 'True'

        if 'all' in request.args and request.args['all']:
            all_overlays = request.args['all'][0] == 'True'
        elif 'overlay_name' in request.args and request.args['overlay_name']:
            overlay_name = request.args['overlay_name'][0]
        else:
            request.setResponseCode(http.PRECONDITION_FAILED)
            return self.twisted_dumps({"success": False, "error": "Parameter 'all' or 'overlay_name' is required"})

        self.enable_overlay_statistics(enable=enable, class_name=overlay_name, all_overlays=all_overlays)
        return self.twisted_dumps({"success": True})

    def enable_overlay_statistics(self, enable=False, class_name=None, all_overlays=False):
        if all_overlays:
            for overlay in self.session.overlays:
                self.session.endpoint.enable_community_statistics(overlay.get_prefix(), enable)
        elif class_name:
            for overlay in self.session.overlays:
                if overlay.__class__.__name__ == class_name:
                    self.session.endpoint.enable_community_statistics(overlay.get_prefix(), enable)



from __future__ import absolute_import

from twisted.trial import unittest

from ..database import Database


class MockDatabase(Database):

    def check_database(self, database_version):
        self.execute("CREATE TABLE option(key TEXT PRIMARY KEY, value BLOB)")
        self.execute("INSERT INTO option(key, value) VALUES('database_version', '0')")
        self.commit()
        return 0


class TestDatabase(unittest.TestCase):

    def setUp(self):
        super(TestDatabase, self).setUp()
        self.database = MockDatabase(u":memory:")

    def test_unloaded(self):
        """
        Check if an unloaded database returns None for queries.
        """
        self.assertIsNone(self.database.execute("SELECT * FROM option"))

    def test_closed(self):
        """
        Check if an unloaded database returns None for queries.
        """
        self.database.open()
        self.assertListEqual([(u'database_version', u'0')], list(self.database.execute("SELECT * FROM option")))
        self.database.close(True)
        self.assertIsNone(self.database.execute("SELECT * FROM option"))



from __future__ import absolute_import

from struct import pack, unpack, unpack_from, calcsize
from socket import inet_aton, inet_ntoa

from ..messaging.payload import Payload
from .routing import Node


def encode_values(values):
    return b''.join([pack('!H', len(value)) + value for value in values])


def decode_values(values_str):
    values = []
    index = 0
    while index < len(values_str):
        length = unpack_from('!H', values_str, offset=index)[0]
        index += calcsize('!H')
        values.append(values_str[index:index + length])
        index += length
    return values


def encode_nodes(nodes):
    nodes_str = b''
    for node in nodes:
        key = node.public_key.key_to_bin()
        nodes_str += inet_aton(node.address[0]) + pack("!H", node.address[1])
        nodes_str += pack('!H', len(key)) + key
    return nodes_str


def decode_nodes(nodes_str):
    nodes = []
    index = 0
    while index < len(nodes_str):
        ip, port, key_length = unpack('!4sHH', nodes_str[index:index + 8])
        index += 8
        address = (inet_ntoa(ip), port)
        key = nodes_str[index:index + key_length]
        index += key_length
        nodes.append(Node(key, address=address))
    return nodes


class BasePayload(Payload):

    format_list = ['I']

    def __init__(self, identifier):
        super(BasePayload, self).__init__()
        self.identifier = identifier

    def to_pack_list(self):
        return [('I', self.identifier)]

    @classmethod
    def from_unpack_list(cls, identifier):
        return BasePayload(identifier)


class PingRequestPayload(BasePayload):
    pass


class PingResponsePayload(BasePayload):
    pass


class StoreRequestPayload(BasePayload):

    format_list = BasePayload.format_list + ['20s', '20s', 'varlenH']

    def __init__(self, identifier, token, target, values):
        super(StoreRequestPayload, self).__init__(identifier)
        self.token = token
        self.target = target
        self.values = values

    def to_pack_list(self):
        data = super(StoreRequestPayload, self).to_pack_list()
        data.append(('20s', self.token))
        data.append(('20s', self.target))
        data.append(('varlenH', encode_values(self.values)))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, token, target, values_str):
        values = decode_values(values_str)
        return StoreRequestPayload(identifier, token, target, values)


class StoreResponsePayload(BasePayload):
    pass


class FindRequestPayload(BasePayload):

    format_list = BasePayload.format_list + ['varlenI', '20s', '?']

    def __init__(self, identifier, lan_address, target, force_nodes):
        super(FindRequestPayload, self).__init__(identifier)
        self.lan_address = lan_address
        self.target = target
        self.force_nodes = force_nodes

    def to_pack_list(self):
        data = super(FindRequestPayload, self).to_pack_list()
        data.append(('varlenI', inet_aton(self.lan_address[0]) + pack("!H", self.lan_address[1])))
        data.append(('20s', self.target))
        data.append(('?', self.force_nodes))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, lan_address, target, force_nodes):
        return FindRequestPayload(identifier,
                                  (inet_ntoa(lan_address[:4]), unpack('!H', lan_address[4:6])[0]),
                                  target,
                                  force_nodes)


class FindResponsePayload(BasePayload):

    format_list = BasePayload.format_list + ['20s', 'varlenH', 'varlenH']

    def __init__(self, identifier, token, values, nodes):
        super(FindResponsePayload, self).__init__(identifier)
        self.token = token
        self.values = values
        self.nodes = nodes

    def to_pack_list(self):
        data = super(FindResponsePayload, self).to_pack_list()
        data.append(('20s', self.token))
        data.append(('varlenH', encode_values(self.values)))
        data.append(('varlenH', encode_nodes(self.nodes)))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, token, values_str, nodes_str):
        return FindResponsePayload(identifier, token, decode_values(values_str), decode_nodes(nodes_str))


class StrPayload(Payload):

    format_list = ['raw']

    def __init__(self, data):
        super(StrPayload, self).__init__()
        self.data = data

    def to_pack_list(self):
        return [('raw', self.data)]

    @classmethod
    def from_unpack_list(cls, data):
        return StrPayload(data)


class SignedStrPayload(Payload):

    format_list = ['varlenH', 'I', 'varlenH']

    def __init__(self, data, version, public_key):
        super(SignedStrPayload, self).__init__()
        self.data = data
        self.version = version
        self.public_key = public_key

    def to_pack_list(self):
        return [('varlenH', self.data),
                ('I', self.version),
                ('varlenH', self.public_key)]

    @classmethod
    def from_unpack_list(cls, data, version, public_key):
        return SignedStrPayload(data, version, public_key)


class StorePeerRequestPayload(BasePayload):

    format_list = BasePayload.format_list + ['20s', '20s']

    def __init__(self, identifier, token, target):
        super(StorePeerRequestPayload, self).__init__(identifier)
        self.token = token
        self.target = target

    def to_pack_list(self):
        data = super(StorePeerRequestPayload, self).to_pack_list()
        data.append(('20s', self.token))
        data.append(('20s', self.target))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, token, target):
        return StorePeerRequestPayload(identifier, token, target)


class StorePeerResponsePayload(BasePayload):
    pass


class ConnectPeerRequestPayload(BasePayload):

    format_list = BasePayload.format_list + ['varlenI', '20s']

    def __init__(self, identifier, lan_address, target):
        super(ConnectPeerRequestPayload, self).__init__(identifier)
        self.lan_address = lan_address
        self.target = target

    def to_pack_list(self):
        data = super(ConnectPeerRequestPayload, self).to_pack_list()
        data.append(('varlenI', inet_aton(self.lan_address[0]) + pack("!H", self.lan_address[1])))
        data.append(('20s', self.target))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, lan_address, target):
        return ConnectPeerRequestPayload(identifier,
                                         (inet_ntoa(lan_address[:4]), unpack('!H', lan_address[4:6])[0]),
                                         target)


class ConnectPeerResponsePayload(BasePayload):

    format_list = BasePayload.format_list + ['varlenH']

    def __init__(self, identifier, nodes):
        super(ConnectPeerResponsePayload, self).__init__(identifier)
        self.nodes = nodes

    def to_pack_list(self):
        data = super(ConnectPeerResponsePayload, self).to_pack_list()
        data.append(('varlenH', encode_nodes(self.nodes)))
        return data

    @classmethod
    def from_unpack_list(cls, identifier, nodes):
        return ConnectPeerResponsePayload(identifier, decode_nodes(nodes))



from __future__ import absolute_import

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyipv8',
    author='Tribler',
    description='The Python implementation of the IPV8 library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.8.0',
    url='https://github.com/Tribler/py-ipv8',
    package_data={'': ['*.*']},
    packages=find_packages(),
    py_modules=['ipv8_service'],
    install_requires=[
        "cryptography",
        "libnacl",
        "netifaces",
        "Twisted",
        "pyOpenSSL",
        "six"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Twisted",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Networking"
    ]
)



from __future__ import absolute_import

from collections import OrderedDict

from six import integer_types, string_types


class OrderedSet(set):
    """
    Looks like a set, iterates like a list.
    """

    def __init__(self, decorated):
        super(OrderedSet, self).__init__(decorated)
        self.decorated = decorated

    def __iter__(self):
        for o in self.decorated:
            yield o

    def __reversed__(self):
        i = len(self.decorated)
        while i > 0:
            i -= 1
            yield self.decorated[i]

    def add(self, element):
        super(OrderedSet, self).add(element)
        self.decorated.append(element)

    def clear(self):
        super(OrderedSet, self).clear()
        self.decorated = []

    def copy(self):
        return OrderedSet([o for o in self.decorated])

    def difference_update(self, s):
        super(OrderedSet, self).difference_update(s)
        self.decorated = [o for o in self.difference(s)]

    def discard(self, element):
        super(OrderedSet, self).discard(element)
        self.decorated.remove(element)

    def intersection_update(self, s):
        super(OrderedSet, self).intersection_update(s)
        self.decorated = [o for o in self.intersection(s)]

    def pop(self):
        if not self.decorated:
            raise KeyError('empty set')
        removed = self.decorated[-1]
        super(OrderedSet, self).remove(removed)
        return removed

    def remove(self, element):
        super(OrderedSet, self).remove(element)
        return self.decorated.remove(element)

    def symmetric_difference_update(self, s):
        super(OrderedSet, self).symmetric_difference_update(s)
        self.decorated = [o for o in self.symmetric_difference(s)]

    def update(self, s):
        super(OrderedSet, self).update(s)
        self.decorated = [o for o in self.union(s)]


class SortableTypeEnum(object):
    """
    Initial type sorting enum: the lower enum values are inserted closer to the head of the list.

    We distinguish:
     - NONE: None
     - BOOLEAN: bool
     - NUMBER: float, int, long
     - STRING: bytes, str, unicode
     - LIST: dict, list, set, tuple*

     * Note that a tuple is a special kind of LIST, which is not sorted.
       This would, for instance, sort (key, value) pairs.
    """

    NONE = 0
    BOOLEAN = 1
    NUMBER = 2
    STRING = 3
    LIST = 4


class Sortable(object):
    """
    Wrapper for sortable types.
    """

    def __init__(self, value):
        """
        Wrap a value as a Sortable, which allows it to be sorted with `sorted()` or `list.sort()`.

        :param value: a wrappable value
        """
        self.source = value  # The source value
        self.value = value  # The sorted value
        if self.value is None:
            self.type = SortableTypeEnum.NONE
        elif isinstance(self.value, bool):
            self.type = SortableTypeEnum.BOOLEAN
        elif isinstance(self.value, (float, integer_types)):
            self.type = SortableTypeEnum.NUMBER
        elif isinstance(self.value, (bytes, string_types)):
            self.type = SortableTypeEnum.STRING
        elif isinstance(self.value, (set, dict, tuple, list)):
            self.type = SortableTypeEnum.LIST
            if isinstance(value, dict):
                self.value = sorted([Sortable((k, v)) for k, v in list(value.items())])
            elif isinstance(value, tuple):
                self.value = [Sortable(v) for v in value]
            else:
                self.value = sorted([Sortable(v) for v in value])
        else:
            raise RuntimeError("Unsortable value %s!" % repr(self.value))

    def compare(self, other):
        """
        Compare this Sortable to another Sortable.

        :param other: the Sortable to compare to
        :type other: Sortable
        :return: -1, 0 or 1 if this object is smaller, equal or larger than the other object
        """
        # Initially we sort on enum value (NONE < BOOLEAN < .. < LIST)
        if other.type == self.type:
            if self.type == SortableTypeEnum.NONE:
                # None is always equal
                return 0
            if self.type == SortableTypeEnum.BOOLEAN:
                # False < True
                if self.value == other.value:
                    return 0
                if self.value:
                    return 1
                return -1
            if self.type == SortableTypeEnum.NUMBER:
                # Python allows us to intrinsically compare float to int and long
                if self.value == other.value:
                    return 0
                if self.value < other.value:
                    return -1
                return 1
            if self.type == SortableTypeEnum.STRING:
                # To deal with unicode we cast all string types to a list of ordinals
                my_ordinals = self.value if isinstance(self.value, bytes) else [ord(c) for c in self.value]
                other_ordinals = other.value if isinstance(other.value, bytes) else [ord(c) for c in other.value]
            else:
                # All list types can simply give their sorted values for comparison
                my_ordinals = self.value
                other_ordinals = other.value
            n = min(len(my_ordinals), len(other_ordinals))
            for i in range(n):
                # Compare each value between lists, this is supported through the Sortable wrapper.
                if my_ordinals[i] < other_ordinals[i]:
                    return -1
                elif my_ordinals[i] > other_ordinals[i]:
                    return 1
            # If all list entries are equal, the shortest list is smaller
            if len(my_ordinals) < len(other_ordinals):
                return -1
            elif len(my_ordinals) > len(other_ordinals):
                return 1
            return 0
        elif self.type < other.type:
            return -1
        return 1

    def finalize(self):
        """
        Unwrap this Sortable to its sorted value.

        Note that LISTs are type-changed:
         - dict -> OrderedDict
         - list -> list
         - set -> list
         - tuple -> tuple (order preserved)

        :return: this Sortable as its source data type
        """
        if self.type == SortableTypeEnum.LIST:
            converted = [v.finalize() for v in self.value]
            if isinstance(self.source, dict):
                out = OrderedDict()
                for v in converted:
                    out[v[0]] = v[1]
                return out
            if isinstance(self.source, set):
                return OrderedSet(converted)
            return tuple(converted) if isinstance(self.source, tuple) else converted
        return self.value

    def __lt__(self, other):
        return self.compare(other) < 0

    def __gt__(self, other):
        return self.compare(other) > 0


def sortable_sort(data):
    """
    Sort complex input containing any combination of:

     - None
     - bool
     - float
     - int
     - long
     - bytes
     - str
     - unicode
     - dict
     - list
     - set
     - tuple

    :return: the sorted input
    """
    return Sortable(data).finalize()


__all__ = ["sortable_sort"]



from __future__ import absolute_import

from twisted.web import http

from .base_endpoint import BaseEndpoint
from ..community import _DEFAULT_ADDRESSES
from ..messaging.anonymization.community import TunnelCommunity
from ..util import cast_to_chr


class IsolationEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for on-demand adding of addresses for different services.

    We support:
     - POST: /isolation?ip=<IP>&port=<PORT>&bootstrapnode=1
     - POST: /isolation?ip=<IP>&port=<PORT>&exitnode=1

    These commands add a bootstrap node and an exit node respectively.
    The IP is a period-seperated string.
    An example call would be:

    curl -X POST "http://localhost:8085/isolation?ip=127.0.0.1&port=9999&bootstrapnode=1"
    """

    def __init__(self, session):
        super(IsolationEndpoint, self).__init__()
        self.session = session

    def add_exit_node(self, address):
        for overlay in self.session.overlays:
            if isinstance(overlay, TunnelCommunity):
                overlay.walk_to(address)

    def add_bootstrap_server(self, address):
        _DEFAULT_ADDRESSES.append(address)
        for overlay in self.session.overlays:
            overlay.walk_to(address)

    def render_POST(self, request):
        # Check if we have arguments, containing an address and the type of address to add.
        if not request.args or b'ip' not in request.args or b'port' not in request.args:
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"success": False, "error": "Parameters 'ip' and 'port' are required"})
        if b'exitnode' not in request.args and b'bootstrapnode' not in request.args:
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"success": False,
                                       "error": "Parameter 'exitnode' or 'bootstrapnode' is required"})
        # Attempt to decode the address
        try:
            address_str = cast_to_chr(request.args[b'ip'][0])
            port_str = cast_to_chr(request.args[b'port'][0])
            fmt_address = (address_str, int(port_str))
        except:
            import traceback
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"success": False, "error": traceback.format_exc()})
        # Actually add the address to the requested service
        if b'exitnode' in request.args:
            self.add_exit_node(fmt_address)
        else:
            self.add_bootstrap_server(fmt_address)
        return self.twisted_dumps({"success": True})



from __future__ import absolute_import

import logging
from base64 import b64decode, b64encode
from binascii import hexlify, unhexlify
from hashlib import sha1

from twisted.internet import reactor
from twisted.internet.task import LoopingCall, deferLater
from twisted.web import http
from twisted.web.server import NOT_DONE_YET

from .base_endpoint import BaseEndpoint
from ..attestation.trustchain.community import TrustChainCommunity
from ..attestation.trustchain.listener import BlockListener
from ..attestation.trustchain.payload import DHTBlockPayload
from ..dht.community import DHTCommunity, MAX_ENTRY_SIZE
from ..dht.discovery import DHTDiscoveryCommunity
from ..keyvault.public.libnaclkey import LibNaCLPK
from ..messaging.serialization import PackError, Serializer

# The period of the block maintenance looping call
BLOCK_REFRESH_PERIOD = 60 * 60  # 1 hour


class DHTEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT data.
    """

    def __init__(self, session):
        super(DHTEndpoint, self).__init__()
        dht_overlays = [overlay for overlay in session.overlays if isinstance(overlay, DHTCommunity)]
        tc_overlays = [overlay for overlay in session.overlays if isinstance(overlay, TrustChainCommunity)]
        if dht_overlays:
            self.putChild(b"statistics", DHTStatisticsEndpoint(dht_overlays[0]))
            self.putChild(b"values", DHTValuesEndpoint(dht_overlays[0]))
            self.putChild(b"peers", DHTPeersEndpoint(dht_overlays[0]))
        if dht_overlays and tc_overlays:
            self.putChild(b"block", DHTBlockEndpoint(dht_overlays[0], tc_overlays[0]))


class DHTBlockEndpoint(BaseEndpoint, BlockListener):
    """
    This endpoint is responsible for returning the latest Trustchain block of a peer. Additionally, it ensures
    this peer's latest TC block is available
    """

    KEY_SUFFIX = b'_BLOCK'
    CHUNK_SIZE = MAX_ENTRY_SIZE - DHTBlockPayload.PREAMBLE_OVERHEAD - 1
    ATTEMPT_LIMIT = 3

    def received_block(self, block):
        """
        Wrapper callback method, inherited from the BlockListener abstract class, which will publish the latest
        TrustChain block to the DHT

        :param block: the latest block added to the Database. This is not actually used by the inner method
        """
        # Check if the block is not null, and if it belongs to this peer
        if block and block.public_key == self.trustchain.my_peer.public_key.key_to_bin():
            deferLater(reactor, 0, self.publish_block, block)

    def should_sign(self, block):
        pass

    def __init__(self, dht, trustchain):
        super(DHTBlockEndpoint, self).__init__()
        self.dht = dht
        self.trustchain = trustchain
        self.block_version = 0
        self.serializer = Serializer()

        self._hashed_dht_key = sha1(self.trustchain.my_peer.public_key.key_to_bin() + self.KEY_SUFFIX).digest()

        # Register a LoopingCall, get it's deferred, and start it.
        self.block_maintenance_task = self.trustchain.register_task('block_maintenance', LoopingCall(
            self.publish_latest_block))
        self.block_maintenance_task.start(BLOCK_REFRESH_PERIOD, now=False)

        trustchain.add_listener(self, [trustchain.UNIVERSAL_BLOCK_LISTENER])

    def publish_latest_block(self):
        """
        Republish the latest TrustChain block under this peer's key
        """
        block = self.trustchain.persistence.get_latest(self.trustchain.my_peer.public_key.key_to_bin())
        if block:
            self.publish_block(block, republish=True)

    def reconstruct_all_blocks(self, block_chunks, public_key):
        """
        Given a list of block chunks, reconstruct all the blocks in a dictionary indexed by their version

        :param block_chunks: the list of block chunks
        :param public_key: the public key of the publishing node, which will be used for verifying the chunks
        :return: a dictionary of block_maintenance_taskreconstructed blocks (in packed format), indexed by the version
                 of the blocks and the maximal version
        """
        new_blocks = {}
        max_version = 0

        for entry in block_chunks:
            try:
                package = self.serializer.unpack_to_serializables([DHTBlockPayload, ], entry)[0]

                pre_signed_content = self.serializer.pack_multiple([('H', package.version),
                                                                    ('H', package.block_position),
                                                                    ('H', package.block_count),
                                                                    ('raw', package.payload)])[0]

                if public_key.verify(package.signature, pre_signed_content):
                    max_version = max_version if max_version > package.version else package.version

                    if package.version not in new_blocks:
                        new_blocks[package.version] = [b''] * package.block_count

                    new_blocks[package.version][package.block_position] = package.payload
            except PackError:
                logging.error("PackError: Found a clandestine entry in the DHT when reconstructing TC blocks: %s",
                              entry)

        # Concatenate the blocks
        for version in new_blocks:
            new_blocks[version] = b''.join(new_blocks[version])

        return new_blocks, max_version

    def publish_block(self, block, republish=False):
        """
        Publishes a block to the DHT, by splitting it in chunks.

        :param block: the block to be published to the DHT
        :param republish: boolean which indicates whether the published block has already been published again
        """
        if block:
            latest_block = block.pack()

            # Get the total number of chunks in this blocks
            total_chunks = len(latest_block) // self.CHUNK_SIZE
            total_chunks += 1 if len(latest_block) % self.CHUNK_SIZE != 0 else 0

            # The actual version of the block may vary based on whether the block is being published or republished
            actual_version = max(self.block_version if not republish else (self.block_version - 1), 0)

            def publish_chunk(_, slice_pointer, chunk_idx, chunk_attempt=1):
                # If we've reached the end of the block, we stop the chain, and increment the block version
                if chunk_idx >= total_chunks:
                    if not republish:
                        self.block_maintenance_task.reset()
                        self.block_version += 1
                    return

                # If we've attempted to publish this block more times than allowed, give up on the chunk and the block
                if chunk_attempt > self.ATTEMPT_LIMIT:
                    logging.error("Publishing latest block failed after %d attempts on chunk %d", self.ATTEMPT_LIMIT,
                                  chunk_idx)
                    return

                # Prepare and pack the chunk for publishing to the DHT
                my_private_key = self.trustchain.my_peer.key
                chunk = latest_block[slice_pointer: slice_pointer + self.CHUNK_SIZE]

                pre_signed_content = self.serializer.pack_multiple([('H', actual_version), ('H', chunk_idx),
                                                                    ('H', total_chunks), ('raw', chunk)])[0]

                signature = my_private_key.signature(pre_signed_content)
                blob_chunk = self.serializer.pack_multiple(
                    DHTBlockPayload(signature, actual_version, chunk_idx, total_chunks, chunk).to_pack_list())

                # Try to add the current chunk to the DHT; if it works, move to the next, otherwise retry
                d = self.dht.store_value(self._hashed_dht_key, blob_chunk[0])
                d.addCallback(publish_chunk, slice_pointer + self.CHUNK_SIZE, chunk_idx + 1, 1)
                d.addErrback(publish_chunk, slice_pointer, chunk_idx, chunk_attempt + 1)

            # On the off chance that the block is actually empty, avoid publishing it
            if total_chunks > 0:
                deferLater(reactor, 0, publish_chunk, None, 0, 0)

    def render_GET(self, request):
        """
        Return the latest TC block of a peer, as identified in the request

        :param request: the request for retrieving the latest TC block of a peer. It must contain the peer's
        public key of the peer
        :return: the latest block of the peer, if found
        """
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        if not request.args or b'public_key' not in request.args:
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"error": "Must specify the peer's public key"})

        def on_success(block_chunks):
            if not block_chunks:
                request.setResponseCode(http.NOT_FOUND)
                request.write(self.twisted_dumps({
                    "error": {
                        "handled": True,
                        "message": "Could not find any blocks for the specified key."
                    }
                }))
            else:
                target_public_key = LibNaCLPK(binarykey=raw_public_key[10:])
                # Discard the 2nd half of the tuples retrieved as a result of the DHT query
                new_blocks, max_version = self.reconstruct_all_blocks([x[0] for x in block_chunks], target_public_key)
                request.write(self.twisted_dumps({"block": b64encode(new_blocks[max_version]).decode('utf-8')}))

            request.finish()

        def on_failure(failure):
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            request.write(self.twisted_dumps({
                "error": {
                    "handled": True,
                    "code": failure.value.__class__.__name__,
                    "message": failure.value.message
                }
            }))
            request.finish()

        raw_public_key = b64decode(request.args[b'public_key'][0])
        hash_key = sha1(raw_public_key + self.KEY_SUFFIX).digest()

        self.dht.find_values(hash_key).addCallbacks(on_success, on_failure)

        return NOT_DONE_YET


class DHTStatisticsEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for returning statistics about the DHT.
    """

    def __init__(self, dht):
        super(DHTStatisticsEndpoint, self).__init__()
        self.dht = dht

    def render_GET(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        buckets = self.dht.routing_table.trie.values()
        stats = {"node_id": hexlify(self.dht.my_node_id),
                 "peer_id": hexlify(self.dht.my_peer.mid),
                 "routing_table_size": sum([len(bucket.nodes) for bucket in buckets]),
                 "routing_table_buckets": len(buckets),
                 "num_keys_in_store": len(self.dht.storage.items),
                 "num_tokens": len(self.dht.tokens)}

        if isinstance(self.dht, DHTDiscoveryCommunity):
            stats.update({
                "num_peers_in_store": {hexlify(key): len(peers) for key, peers in self.dht.store.items()},
                "num_store_for_me": {hexlify(key): len(peers) for key, peers in self.dht.store_for_me.items()}
            })

        return self.twisted_dumps({"statistics": stats})


class DHTPeersEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT peers.
    """

    def __init__(self, dht):
        super(DHTPeersEndpoint, self).__init__()
        self.dht = dht

    def getChild(self, path, request):
        return SpecificDHTPeerEndpoint(self.dht, path)


class SpecificDHTPeerEndpoint(BaseEndpoint):
    """
    This class handles requests for a specific DHT peer.
    """

    def __init__(self, dht, key):
        super(SpecificDHTPeerEndpoint, self).__init__()
        self.mid = bytes(unhexlify(key))
        self.dht = dht

    def render_GET(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        def on_success(nodes):
            node_dicts = []
            for node in nodes:
                node_dicts.append({
                    'public_key': b64encode(node.public_key.key_to_bin()).decode('utf-8'),
                    'address': node.address
                })
            request.write(self.twisted_dumps({"peers": node_dicts}))
            request.finish()

        def on_failure(failure):
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            request.write(self.twisted_dumps({
                "error": {
                    "handled": True,
                    "code": failure.value.__class__.__name__,
                    "message": failure.value.message
                }
            }))
            request.finish()

        self.dht.connect_peer(self.mid).addCallbacks(on_success, on_failure)

        return NOT_DONE_YET


class DHTValuesEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT values.
    """

    def __init__(self, dht):
        super(DHTValuesEndpoint, self).__init__()
        self.dht = dht

    def render_GET(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        results = {}
        for key, raw_values in self.dht.storage.items.items():
            values = self.dht.post_process_values([v.data for v in raw_values])
            dicts = []
            for value in values:
                data, public_key = value
                dicts.append({
                    'public_key': b64encode(public_key).decode('utf-8') if public_key else None,
                    'value': hexlify(data)
                })
            results[hexlify(key)] = dicts

        return self.twisted_dumps(results)

    def getChild(self, path, request):
        return SpecificDHTValueEndpoint(self.dht, path)


class SpecificDHTValueEndpoint(BaseEndpoint):
    """
    This class handles requests for a specific DHT value.
    """

    def __init__(self, dht, key):
        super(SpecificDHTValueEndpoint, self).__init__()
        self.key = bytes(unhexlify(key))
        self.dht = dht

    def render_GET(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        def on_success(values):
            dicts = []
            for value in values:
                data, public_key = value
                dicts.append({
                    'public_key': b64encode(public_key).decode('utf-8') if public_key else None,
                    'value': hexlify(data)
                })
            request.write(self.twisted_dumps({"values": dicts}))
            request.finish()

        def on_failure(failure):
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            request.write(self.twisted_dumps({
                "error": {
                    "handled": True,
                    "code": failure.value.__class__.__name__,
                    "message": failure.value.message
                }
            }))
            request.finish()

        self.dht.find_values(self.key).addCallbacks(on_success, on_failure)

        return NOT_DONE_YET

    def render_PUT(self, request):
        if not self.dht:
            request.setResponseCode(http.NOT_FOUND)
            return self.twisted_dumps({"error": "DHT community not found"})

        def on_success(_):
            request.write(self.twisted_dumps({"stored": True}))
            request.finish()

        def on_failure(failure):
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            request.write(self.twisted_dumps({
                "error": {
                    "handled": True,
                    "code": failure.value.__class__.__name__,
                    "message": failure.value.message
                }
            }))
            request.finish()

        parameters = http.parse_qs(request.content.read(), 1)
        if 'value' not in parameters:
            request.setResponseCode(http.BAD_REQUEST)
            return self.twisted_dumps({"error": "incorrect parameters"})

        self.dht.store_value(self.key, unhexlify(parameters['value'][0]), sign=True).addCallbacks(on_success,
                                                                                                  on_failure)

        return NOT_DONE_YET



from __future__ import absolute_import

from six.moves import xrange
from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.internet.task import Clock, deferLater, LoopingCall

from .base import TestBase
from ..taskmanager import CLEANUP_FREQUENCY, TaskManager


class TestTaskManager(TestBase):

    def setUp(self):
        super(TestTaskManager, self).setUp()

        self.dispersy_objects = []
        self.tm = TaskManager()
        self.tm._reactor = Clock()

        self.counter = 0

    def tearDown(self):
        self.tm.shutdown_task_manager()
        return super(TestTaskManager, self).tearDown()

    def test_call_later(self):
        self.tm.register_task("test", reactor.callLater(10, lambda: None))

        self.assertTrue(self.tm.is_pending_task_active("test"))

    def test_call_later_and_cancel(self):
        self.tm.register_task("test", reactor.callLater(10, lambda: None))
        self.tm.cancel_pending_task("test")

        self.assertFalse(self.tm.is_pending_task_active("test"))

    def test_call_later_and_replace(self):
        task1 = self.tm.register_task("test", reactor.callLater(10, lambda: None))
        self.tm.replace_task("test", reactor.callLater(10, lambda: None))

        self.assertTrue(self.tm.is_pending_task_active("test"))
        self.assertFalse(task1.active())

    def test_looping_call(self):
        self.tm.register_task("test", LoopingCall(lambda: None)).start(10, now=True)

        self.assertTrue(self.tm.is_pending_task_active("test"))

    def test_looping_call_and_cancel(self):
        self.tm.register_task("test", LoopingCall(lambda: None)).start(10, now=True)
        self.tm.cancel_pending_task("test")

        self.assertFalse(self.tm.is_pending_task_active("test"))

    def test_delayed_looping_call_requires_interval(self):
        self.assertRaises(ValueError, self.tm.register_task, "test", LoopingCall(lambda: None), delay=1)

    def test_delayed_looping_call_requires_LoopingCall_or_Deferred(self):
        self.assertRaises(ValueError, self.tm.register_task, "test not Deferred nor LoopingCall",
                          self.tm._reactor.callLater(0, lambda: None), delay=1)

    def test_delayed_looping_call_register_and_cancel_pre_delay(self):
        self.assertFalse(self.tm.is_pending_task_active("test"))
        self.tm.register_task("test", LoopingCall(lambda: None), delay=1, interval=1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))

    def test_delayed_looping_call_register_wait_and_cancel(self):
        self.assertFalse(self.tm.is_pending_task_active("test"))
        lc = LoopingCall(self.count)
        lc.clock = self.tm._reactor
        self.tm.register_task("test", lc, delay=1, interval=1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        # After one second, the counter has increased by one and the task is still active.
        self.tm._reactor.advance(1)
        self.assertEquals(1, self.counter)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        # After one more second, the counter should be 2
        self.tm._reactor.advance(1)
        self.assertEquals(2, self.counter)
        # After canceling the task the counter should stop increasing
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))
        self.tm._reactor.advance(10)
        self.assertEquals(2, self.counter)

    def test_delayed_deferred(self):
        self.assertFalse(self.tm.is_pending_task_active("test"))
        d = Deferred()
        d.addCallback(self.set_counter)
        self.tm.register_task("test", d, delay=1, value=42)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        # After one second, the deferred has fired
        self.tm._reactor.advance(1)
        self.assertEquals(42, self.counter)
        self.assertFalse(self.tm.is_pending_task_active("test"))

    def test_raise_on_duplicate_task_name(self):
        self.tm.register_task("test", reactor.callLater(10, lambda: None))
        with self.assertRaises(RuntimeError):
            self.tm.register_task("test", reactor.callLater(10, lambda: None))

    def test_duplicate_anon_task_name(self):
        self.tm.register_anonymous_task("test", deferLater(reactor, 10, lambda: None)).addErrback(lambda _: None)
        self.tm.register_anonymous_task("test", deferLater(reactor, 10, lambda: None)).addErrback(lambda _: None)

        deferred_list = self.tm.wait_for_deferred_tasks()
        self.assertEqual(2, len(deferred_list.resultList))

        self.tm.cancel_all_pending_tasks()

    def test_duplicate_anon_task_deferred(self):
        task = deferLater(reactor, 10, lambda: None)
        self.tm.register_anonymous_task("test", task).addErrback(lambda _: None)
        with self.assertRaises(RuntimeError):
            self.tm.register_anonymous_task("test", task).addErrback(lambda _: None)

        self.tm.cancel_all_pending_tasks()

    def test_shutdown(self):
        """
        Test if the TaskManager does not allow new tasks after shutdown().
        """
        self.tm.shutdown_task_manager()
        self.tm.register_anonymous_task("test", deferLater(reactor, 10, lambda: None)).addErrback(lambda _: None)

        deferred_list = self.tm.wait_for_deferred_tasks()
        self.assertEqual(0, len(deferred_list.resultList))

    def test_cleanup(self):
        """
        Test if the tasks are cleaned up after the cleanup frequency has been met.
        """
        deferred = succeed(None)
        for _ in xrange(CLEANUP_FREQUENCY):
            self.tm.register_anonymous_task("test", deferred)

        deferred_list = self.tm.wait_for_deferred_tasks()
        self.assertEqual(0, len(deferred_list.resultList))

    def test_cleanup_remaining(self):
        """
        Test if tasks which have yet to complete are not cleaned.
        """
        deferred = succeed(None)
        self.tm.register_anonymous_task("test", Deferred()).addErrback(lambda _: None)
        for _ in xrange(CLEANUP_FREQUENCY - 1):
            self.tm.register_anonymous_task("test", deferred)

        deferred_list = self.tm.wait_for_deferred_tasks()
        self.assertEqual(1, len(deferred_list.resultList))

    def count(self):
        self.counter += 1

    def set_counter(self, value):
        self.counter = value



from __future__ import absolute_import

import logging
from threading import RLock

from twisted.internet import reactor
from twisted.internet.base import DelayedCall
from twisted.internet.defer import Deferred, DeferredList
from twisted.internet.task import LoopingCall

CLEANUP_FREQUENCY = 100


class TaskManager(object):

    """
    Provides a set of tools to mantain a list of twisted "tasks" (Deferred, LoopingCall, DelayedCall) that are to be
    executed during the lifetime of an arbitrary object, usually getting killed with it.
    """
    _reactor = reactor

    def __init__(self):
        self._pending_tasks = {}
        self._cleanup_counter = CLEANUP_FREQUENCY
        self._task_lock = RLock()
        self._shutdown = False
        self._logger = logging.getLogger(self.__class__.__name__)

    def replace_task(self, name, task):
        """
        Replace named task with the new one, cancelling the old one in the process.
        """
        with self._task_lock:
            self.cancel_pending_task(name)
            return self.register_task(name, task)

    def register_task(self, name, task, delay=None, value=None, interval=None):
        """
        Register a task so it can be canceled at shutdown time or by name.
        """
        with self._task_lock:
            if self._shutdown:
                self._logger.warning("Not adding task %s due to shutdown!", str(task))
                is_active, stopfn = self._get_isactive_stopper(task)
                if is_active and stopfn:
                    stopfn()
                return task

            assert isinstance(task, (Deferred, DelayedCall, LoopingCall)), (task, isinstance(task, Deferred))

            if self.is_pending_task_active(name):
                self.replace_task(name, task)
                raise RuntimeError("Task already exists: '%s'" % name)

            if delay is not None:
                if isinstance(task, Deferred):
                    dc = self._reactor.callLater(delay, task.callback, value)
                elif isinstance(task, LoopingCall):
                    if interval is None:
                        raise ValueError("Expecting interval for delayed LoopingCall")
                    dc = self._reactor.callLater(delay, task.start, interval)
                else:
                    raise ValueError("Expecting Deferred or LoopingCall if task is delayed")

                task = (dc, task)

            self._maybe_clean_task_list()
            self._pending_tasks[name] = task
            return task

    def register_anonymous_task(self, basename, task, delay=None, value=None, interval=None):
        """
        Wrapper for register_task to derive a unique name from the basename.
        """
        return self.register_task(basename + str(id(task)), task, delay, value, interval)

    def cancel_pending_task(self, name):
        """
        Cancels the named task
        """
        with self._task_lock:
            self._maybe_clean_task_list()

            task = self._pending_tasks.get(name, None)
            if not task:
                return

            is_active, stopfn = self._get_isactive_stopper(task)
            if is_active and stopfn:
                stopfn()
                self._pending_tasks.pop(name, None)

    def cancel_all_pending_tasks(self):
        """
        Cancels all the registered tasks.
        This usually should be called when stopping or destroying the object so no tasks are left floating around.
        """
        with self._task_lock:
            assert all([isinstance(task, (Deferred, DelayedCall, LoopingCall, tuple))
                        for task in self._pending_tasks.values()]), self._pending_tasks

            for name in list(self._pending_tasks.keys()):
                self.cancel_pending_task(name)

    def is_pending_task_active(self, name):
        """
        Return a boolean determining if a task is active.
        """
        with self._task_lock:
            task = self._pending_tasks.get(name, None)
            return self._get_isactive_stopper(task)[0] if task else False

    def wait_for_deferred_tasks(self):
        """
        Returns a deferred that will fire when all registered Deferreds are done.
        """
        with self._task_lock:
            self._maybe_clean_task_list()
            return DeferredList(list(self._iter_deferreds()))

    def _iter_deferreds(self):
        with self._task_lock:
            for task in self._pending_tasks.values():
                if isinstance(task, Deferred):
                    yield task

    def _get_isactive_stopper(self, task):
        """
        Return a boolean determining if a task is active and its cancel/stop method if the task is registered.
        """
        with self._task_lock:
            if isinstance(task, Deferred):
                # Have in mind that any deferred in the pending tasks list should have been constructed with a
                # canceller function.
                return not task.called, getattr(task, 'cancel', None)
            elif isinstance(task, DelayedCall):
                return task.active(), task.cancel
            elif isinstance(task, LoopingCall):
                return task.running, task.stop
            elif isinstance(task, tuple):
                if task[0].active():
                    return task[0].active(), task[0].cancel
                else:
                    return self._get_isactive_stopper(task[1])
            else:
                return False, None

    def _maybe_clean_task_list(self):
        """
        Removes finished tasks from the task list.
        """
        with self._task_lock:
            if self._cleanup_counter:
                self._cleanup_counter -= 1
            else:
                self._cleanup_counter = CLEANUP_FREQUENCY
                for name in list(self._pending_tasks.keys()):
                    if not self.is_pending_task_active(name):
                        self._pending_tasks.pop(name, None)

    def shutdown_task_manager(self):
        """
        Clear the task manager, cancel all pending tasks and disallow new tasks being added.
        """
        with self._task_lock:
            self._shutdown = True
            self.cancel_all_pending_tasks()


__all__ = ["TaskManager"]



from __future__ import absolute_import

from twisted.internet import reactor
from twisted.internet.defer import fail, inlineCallbacks, returnValue, succeed
from twisted.internet.task import deferLater
from twisted.internet.threads import deferToThread

from .base import TestBase
from ..util import blocking_call_on_reactor_thread


class TestUtil(TestBase):

    @inlineCallbacks
    def test_blocking_call_in_thread(self):
        """
        Check if the reactor thread is properly blocked by a threaded function.
        """
        @blocking_call_on_reactor_thread
        @inlineCallbacks
        def waiter():
            # 'Release' our claim on the reactor thread.
            # blocking_call_on_reactor_thread should prevent anything else being scheduled though.
            yield deferLater(reactor, 0.01, lambda: None)
            waiter.variable += 1
            returnValue(waiter.variable)

        @blocking_call_on_reactor_thread
        def quicker():
            # Immediately use the reactor thread and return
            waiter.variable += 1
            return succeed(waiter.variable)

        waiter.variable = 1

        # 'Release' the reactor thread and increment waiter.variable
        # If release didn't allow other to be scheduled, waiter.variable is now 2
        # If quicker() came first, waiter.variable is now 3 (bad)
        value = yield deferToThread(waiter)
        # Claim reactor thread and increment waiter.variable
        # If waiter() came first, waiter.variable is now 3
        # If quicker() managed to sneak in before this, waiter.variable is now 2 (bad)
        value2 = yield deferToThread(quicker)

        self.assertEqual(value, 2)
        self.assertEqual(value2, 3)

    @inlineCallbacks
    def test_blocking_call_in_thread_with_error(self):
        """
        Check if a blocking call propagates its errors from a thread.
        """
        @blocking_call_on_reactor_thread
        def quicker():
            return fail(RuntimeError())

        success = True
        try:
            yield deferToThread(quicker)
        except RuntimeError:
            success = False

        self.assertFalse(success)



from __future__ import absolute_import
from __future__ import division

import binascii
import random
import time

from collections import deque
from threading import RLock
from socket import inet_aton

from ..peer import Peer
from .trie import Trie
from ..util import cast_to_bin, cast_to_unicode

# By default we allow a maximum number of 10 queries during a 5s interval.
# Additional queries will be dropped.
NODE_LIMIT_INTERVAL = 5
NODE_LIMIT_QUERIES = 10

NODE_STATUS_GOOD = 2
NODE_STATUS_UNKNOWN = 1
NODE_STATUS_BAD = 0

MAX_BUCKET_SIZE = 20


def id_to_binary_string(node_id):
    return format(int(binascii.hexlify(cast_to_bin(node_id)), 16), '0160b')


def distance(a, b):
    return int(binascii.hexlify(cast_to_bin(a)), 16) ^ int(binascii.hexlify(cast_to_bin(b)), 16)


def calc_node_id(ip, mid):
    # Loosely based on the Bittorrent DHT (https://libtorrent.org/dht_sec.html), the node id is calculated as follows:
    # first 3 bytes of crc32c(ip & 0x030f3fff) + first 17 bytes of sha1(public_key)
    ip_bin = inet_aton(ip)
    ip_mask = '\x03\x0f\x3f\0xff'
    ip_masked = ''.join([chr(ord(ip_bin[i:i + 1]) & ord(ip_mask[i:i + 1])) for i in range(4)])

    crc32_unsigned = binascii.crc32(cast_to_bin(ip_masked)) % (2 ** 32)
    crc32_bin = binascii.unhexlify('%08x' % crc32_unsigned)

    return crc32_bin[:3] + mid[:17]


class Node(Peer):
    """
    The Node class represents a peer within the DHT community
    """

    def __init__(self, *args, **kwargs):
        super(Node, self).__init__(*args, **kwargs)
        self.bucket = None
        self.last_response = 0
        self.last_queries = deque(maxlen=NODE_LIMIT_QUERIES)
        self.failed = 0
        self.rtt = 0

    @property
    def id(self):
        return calc_node_id(self.address[0], self.mid)

    @property
    def last_contact(self):
        return max(self.last_response, self.last_query)

    @property
    def last_query(self):
        return self.last_queries[-1] if self.last_queries else 0

    @property
    def blocked(self):
        if len(self.last_queries) < self.last_queries.maxlen:
            return False
        return time.time() - self.last_queries[0] < NODE_LIMIT_INTERVAL

    @property
    def status(self):
        # A good node is a node has responded to one of our queries within the last 15 minutes, or has ever responded
        # to one of our queries and has sent us a query within the last 15 minutes. This is the same logic as
        # used in BEP-5
        now = time.time()
        if self.failed >= 2:
            return NODE_STATUS_BAD
        elif ((now - self.last_response) < 15 * 60) or (self.last_response > 0 and (now - self.last_query) < 15 * 60):
            return NODE_STATUS_GOOD
        return NODE_STATUS_UNKNOWN

    def distance(self, other_node):
        return distance(self.id, other_node.id)


class Bucket(object):
    """
    The Bucket class stores nodes that share common prefix ID.
    """

    def __init__(self, prefix_id, max_size=MAX_BUCKET_SIZE):
        self.nodes = {}
        self.prefix_id = prefix_id
        self.max_size = max_size
        self.last_changed = 0

    def generate_id(self):
        rand_node_id_bin = format(random.randint(0, 2 ** (160 - len(self.prefix_id))), '0160b')
        return binascii.unhexlify(format(int(rand_node_id_bin, 2), '040X'))

    def owns(self, node_id):
        node_id_binary = id_to_binary_string(node_id)
        return node_id_binary.startswith(self.prefix_id)

    def get(self, node_id):
        return self.nodes.get(node_id)

    def add(self, node):
        # Is this node allowed to be in this bucket?
        if not self.owns(node.id):
            return False

        # Update existing node
        elif node.id in self.nodes:
            curr_node = self.nodes[node.id]
            curr_node.address = node.address
            self.last_changed = time.time()
            return True

        # Make room if needed
        if len(self.nodes) >= self.max_size:
            for n in list(self.nodes.values()):
                if n.status == NODE_STATUS_BAD:
                    del self.nodes[n.id]
                    break

            for n in list(self.nodes.values()):
                if node.rtt and n.rtt / node.rtt >= 2.0:
                    del self.nodes[n.id]
                    break

        # Insert
        if len(self.nodes) < self.max_size:
            self.nodes[node.id] = node
            node.bucket = self
            self.last_changed = time.time()
            return True

        return False

    def split(self):
        if len(self.nodes) < self.max_size:
            return False

        b_0 = Bucket(self.prefix_id + u'0', self.max_size)
        b_1 = Bucket(self.prefix_id + u'1', self.max_size)
        for node in list(self.nodes.values()):
            if b_0.owns(node.id):
                b_0.add(node)
            elif b_1.owns(node.id):
                b_1.add(node)
            else:
                import logging
                logging.error('Failed to place node into bucket while splitting')
        return b_0, b_1


class RoutingTable(object):
    """
    The RoutingTable is a binary tree that keeps track of Nodes that we have a connection to.
    """

    def __init__(self, my_node_id):
        self.my_node_id = my_node_id
        self.trie = Trie(u'01')
        self.trie[u''] = Bucket(u'')
        self.lock = RLock()

    def get_bucket(self, node_id):
        node_id_binary = id_to_binary_string(node_id)
        return self.trie.longest_prefix_value(cast_to_unicode(node_id_binary), default=None) or self.trie[u'']

    def add(self, node):
        with self.lock:
            bucket = self.get_bucket(node.id)

            # Add/update node
            if not bucket.add(node):
                # If adding the node failed, split the bucket
                # Splitting is only allowed if our own node_id falls within this bucket
                if bucket.owns(self.my_node_id):
                    bucket_0, bucket_1 = bucket.split()
                    self.trie[bucket.prefix_id + u'0'] = bucket_0
                    self.trie[bucket.prefix_id + u'1'] = bucket_1
                    del self.trie[bucket.prefix_id]

                    # Retry
                    return self.add(node)
            else:
                return bucket.get(node.id)

    def remove_bad_nodes(self):
        with self.lock:
            for bucket in self.trie.values():
                for node_id, node in bucket.nodes.items():
                    if node.status == NODE_STATUS_BAD:
                        bucket.nodes.pop(node_id, None)

    def has(self, node_id):
        return bool(self.get(node_id))

    def get(self, node_id):
        return self.get_bucket(node_id).get(node_id)

    def closest_nodes(self, node_id, max_nodes=8, exclude_node=None):
        with self.lock:
            hash_binary = cast_to_unicode(id_to_binary_string(node_id))
            prefix = self.trie.longest_prefix(hash_binary, default=u'')

            nodes = set()
            for i in reversed(range(len(prefix) + 1)):
                for suffix in self.trie.suffixes(prefix[:i]):
                    bucket = self.trie[prefix[:i] + suffix]
                    nodes |= {node for node in list(bucket.nodes.values())
                              if node.status != NODE_STATUS_BAD and (exclude_node is None
                                                                     or node.id != exclude_node.id)}

                # Limit number of nodes returned
                if len(nodes) > max_nodes:
                    break

            # Ensure nodes are sorted by distance
            return sorted(nodes, key=lambda n: (distance(n.id, node_id), n.status))[:max_nodes]



from __future__ import absolute_import

from twisted.internet.defer import inlineCallbacks, returnValue

from .peer_communication import IGetStyleRequestsAE, IPostStyleRequestsAE
from ...mocking.rest.rest_peer_communication import HTTPRequester, RequestException, process_json_response


class HTTPGetRequesterAE(IGetStyleRequestsAE, HTTPRequester):
    """
    Implements the GetStyleRequests abstract methods using the HTTP protocol for the attestation endpoint.
    """

    def __init__(self):
        IGetStyleRequestsAE.__init__(self)
        HTTPRequester.__init__(self)

    @process_json_response
    @inlineCallbacks
    def make_outstanding(self, param_dict):
        """
        Forward a request for outstanding attestation requests.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           {'type': 'outstanding'},
                                           param_dict.get('callback', None))
        returnValue(response)

    @process_json_response
    @inlineCallbacks
    def make_verification_output(self, param_dict):
        """
        Forward a request for the verification outputs.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           {'type': 'verification_output'},
                                           param_dict.get('callback', None))
        returnValue(response)

    @process_json_response
    @inlineCallbacks
    def make_peers(self, param_dict):
        """
        Forward a request for the known peers in the network.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           {'type': 'peers'},
                                           param_dict.get('callback', None))
        returnValue(response)

    @process_json_response
    @inlineCallbacks
    def make_attributes(self, param_dict):
        """
        Forward a request for the attributes of a peer.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (attributes), and the (optional) b64_mid of the attester
        request_parameters = param_dict.get('request_parameters', dict())
        request_parameters.update({'type': 'attributes'})

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)

    @process_json_response
    @inlineCallbacks
    def make_dht_block(self, param_dict):
        """
        Forward a request for the latest TC block of a peer

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                'public_key': the public key of the peer whose latest TC block is being requested
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: None
        :raises RequestException: raised when the method could not find one of the required pieces of information
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)
        request_parameters = {}

        if 'public_key' in param_dict:
            request_parameters['public_key'] = param_dict['public_key']
        else:
            raise RequestException("Malformed request: did not specify the public_key")

        response = yield self.make_request("http://{0}:{1}/{2}".format(interface, port, endpoint),
                                           'GET',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)

    @inlineCallbacks
    def make_drop_identity(self, param_dict):
        """
        Forward a request for dropping a peer's identity.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           {'type': 'drop_identity'},
                                           param_dict.get('callback', None))
        returnValue(response)

    @process_json_response
    @inlineCallbacks
    def make_outstanding_verify(self, param_dict):
        """
        Forward a request which requests information on the outstanding verify requests

        :param param_dict: Should have at least the following structure:
            {
                    'interface': target peer IP or alias
                    'port': port_number
                    'endpoint': endpoint_name
                    (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (request), and the rest of the parameters
        request_parameters = {'type': 'outstanding_verify'}

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'GET',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)


class HTTPPostRequesterAE(IPostStyleRequestsAE, HTTPRequester):
    """
    Implements the PostStyleRequests abstract methods using the HTTP protocol for the AttestationEndpoint
    """

    def __init__(self):
        IPostStyleRequestsAE.__init__(self)
        HTTPRequester.__init__(self)

    @inlineCallbacks
    def make_attestation_request(self, param_dict):
        """
        Forward a request for the attestation of an attribute.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                'attribute_name': attribute_name
                'mid': attester b64_mid
                (optional) 'metadata': JSON style metadata required for the attestation process
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (request), and the rest of the parameters
        request_parameters = {'type': 'request', 'id_format': 'id_metadata'}

        # Add the request parameters one-by-one; if required parameter is missing, then raise error
        if 'attribute_name' in param_dict:
            request_parameters['attribute_name'] = param_dict['attribute_name']
        else:
            raise RequestException("Malformed request: did not specify the attribute_name")

        if 'mid' in param_dict:
            request_parameters['mid'] = param_dict['mid']
        else:
            raise RequestException("Malformed request: did not specify the attester's mid")

        if 'metadata' in param_dict:
            request_parameters['metadata'] = param_dict['metadata']

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'POST',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)

    @inlineCallbacks
    def make_attest(self, param_dict):
        """
        Forward a request which attests an attestation request.

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                'attribute_name': attribute_name
                'mid': attestee's b64_mid
                'attribute_value': b64 hash of the attestation blob
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (attest), and the rest of the parameters
        request_parameters = {'type': 'attest'}

        # Add the request parameters one-by-one; if required parameter is missing, then raise error
        if 'attribute_name' in param_dict:
            request_parameters['attribute_name'] = param_dict['attribute_name']
        else:
            raise RequestException("Malformed request: did not specify the attribute_name")

        if 'mid' in param_dict:
            request_parameters['mid'] = param_dict['mid']
        else:
            raise RequestException("Malformed request: did not specify the attestee's mid")

        if 'attribute_value' in param_dict:
            request_parameters['attribute_value'] = param_dict['attribute_value']
        else:
            raise RequestException("Malformed request: did not specify the attribute_value, i.e. the attestation"
                                   "blob hash")

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'POST',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)

    @inlineCallbacks
    def make_verify(self, param_dict):
        """
        Forward a request which demands the verification of an attestation

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                'attribute_hash': the b64 hash of the attestation blob which needs to be verified
                'mid': verifier's b64_mid
                'attribute_values': a string of b64 encoded values, which are separated by ',' characters
                                    e.g. "val_1,val_2,val_3, ..., val_N"
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (attest), and the rest of the parameters
        request_parameters = {'type': 'verify', 'id_format': 'id_metadata'}

        # Add the request parameters one-by-one; if required parameter is missing, then raise error
        if 'attribute_hash' in param_dict:
            request_parameters['attribute_hash'] = param_dict['attribute_hash']
        else:
            raise RequestException("Malformed request: did not specify the attribute_hash")

        if 'mid' in param_dict:
            request_parameters['mid'] = param_dict['mid']
        else:
            raise RequestException("Malformed request: did not specify the verifier's mid")

        if 'attribute_values' in param_dict:
            request_parameters['attribute_values'] = param_dict['attribute_values']
        else:
            raise RequestException("Malformed request: did not specify the attribute_values")

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'POST',
                                           request_parameters,
                                           param_dict.get('callback', None))

        returnValue(response)

    @inlineCallbacks
    def make_allow_verify(self, param_dict):
        """
        Forward a request which requests that verifications be allowed for a particular peer for a particular attribute

        :param param_dict: Should have at least the following structure:
            {
                'interface': target peer IP or alias
                'port': port_number
                'endpoint': endpoint_name
                'attribute_name': attribute_name
                'mid': verifier's b64_mid
                (optional) 'callback': single parameter callback for the request's response
            }
        :return: the request's response
        :raises RequestException: raised when the method could not find some element required for the construction of
                                  the request
        """
        interface, port, endpoint = HTTPRequester.get_access_parameters(param_dict)

        # Add the type of the request (request), and the rest of the parameters
        request_parameters = {'type': 'allow_verify'}

        # Add the request parameters one-by-one; if required parameter is missing, then raise error
        if 'attribute_name' in param_dict:
            request_parameters['attribute_name'] = param_dict['attribute_name']
        else:
            raise RequestException("Malformed request: did not specify the attribute_name")

        if 'mid' in param_dict:
            request_parameters['mid'] = param_dict['mid']
        else:
            raise RequestException("Malformed request: did not specify the attester's mid")

        response = yield self.make_request(HTTPRequester.basic_url_builder(interface, port, endpoint),
                                           'POST',
                                           request_parameters,
                                           param_dict.get('callback', None))
        returnValue(response)



from __future__ import absolute_import
from __future__ import print_function

from os import chdir, getcwd, mkdir, path
from random import randint
import time

from twisted.internet import reactor

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import sys
    sys.path.append(path.abspath(path.join(path.dirname(__file__), "..")))

from ipv8_service import _COMMUNITIES, IPv8, _WALKERS  # pylint: disable=ungrouped-imports
from ipv8.configuration import get_default_configuration  # pylint: disable=ungrouped-imports
from ipv8.keyvault.crypto import ECCrypto  # pylint: disable=ungrouped-imports
from ipv8.peer import Peer  # pylint: disable=ungrouped-imports


START_TIME = time.time()
RESULTS = {}


# Wait until we get a non-tracker response
# Once all overlays have finished, stop the script
def custom_intro_response_cb(self, peer, dist, payload):
    if (peer.address not in self.network.blacklist) and (self.__class__.__name__ not in RESULTS):
        RESULTS[self.__class__.__name__] = time.time() - START_TIME
        print(self.__class__.__name__, "found a peer!", file=sys.stderr)
        if len(get_default_configuration()['overlays']) == len(RESULTS):
            reactor.callFromThread(reactor.stop)


# If it takes longer than 30 seconds to find anything, abort the experiment and set the intro time to -1.0
def on_timeout():
    for definition in get_default_configuration()['overlays']:
        if definition['class'] not in RESULTS:
            RESULTS[definition['class']] = -1.0
            print(definition['class'], "found no peers at all!", file=sys.stderr)
    reactor.callFromThread(reactor.stop)


# Override the Community master peers so we don't interfere with the live network
# Also hook in our custom logic for introduction responses
for community_cls in _COMMUNITIES.values():
    community_cls.master_peer = Peer(ECCrypto().generate_key(u"medium"))
    community_cls.introduction_response_callback = custom_intro_response_cb

# Create two peers with separate working directories
previous_workdir = getcwd()
for i in [1, 2]:
    configuration = get_default_configuration()
    configuration['port'] = 12000 + randint(0, 10000)
    configuration['logger']['level'] = "CRITICAL"
    for overlay in configuration['overlays']:
        overlay['walkers'] = [walker for walker in overlay['walkers'] if walker['strategy'] in _WALKERS]
    workdir = path.abspath(path.join(path.dirname(__file__), "%d" % i))
    if not path.exists(workdir):
        mkdir(workdir)
    chdir(workdir)
    IPv8(configuration)
    chdir(previous_workdir)

# Actually start running everything, this blocks until the experiment finishes
reactor.callLater(30.0, on_timeout)
reactor.run()

# Print the introduction times for all default Communities, sorted alphabetically.
print(','.join(['%.4f' % RESULTS[key] for key in sorted(RESULTS)]))



from __future__ import absolute_import

import struct

from .value import FP2Value


__all__ = ['ipack', 'iunpack', 'pack_pair', 'unpack_pair', 'BonehPublicKey', 'BonehPrivateKey']


def _num_to_str(num):
    """
    Convert an integer to a str.
    """
    out = b''
    h = hex(num)[2:]
    if h.endswith('L'):
        h = h[:-1]
    if (len(h) % 2) == 1:
        h = '0' + h
    for b in range(0, len(h), 2):
        out += struct.pack(">B", int(h[b] + h[b + 1], 16))
    return out


def _str_to_num(s):
    """
    Convert a str to an integer.
    """
    out = 0
    for i in range(len(s)):
        out <<= 8
        out |= struct.unpack(">B", s[i:i + 1])[0]
    return out


def ipack(num):
    """
    Serialize an integer.
    """
    pnum = _num_to_str(num)
    l = _num_to_str(len(pnum))
    return struct.pack(">B", len(l)) + l + pnum


def iunpack(s):
    """
    Unserialize an integer from a str.
    """
    llen = struct.unpack(">B", s[0:1])[0]
    l = _str_to_num(s[1:1 + llen])
    return _str_to_num(s[1 + llen:llen + l + 1]), s[llen + l + 1:]


def pack_pair(a, b):
    """
    Serialize a pair of two integers.
    """
    return ipack(a) + ipack(b)


def unpack_pair(s):
    """
    Unserialize a pair of two integers.
    """
    a, r = iunpack(s)
    b, r = iunpack(r)
    return a, b, r


class BonehPublicKey(object):
    """
    A public key for Boneh et al.'s cryptosystem.
    """
    FIELDS = 5

    def __init__(self, p, g, h):
        self.p = p
        self.g = g
        self.h = h

    def serialize(self):
        return ipack(self.p) + ipack(self.g.a) + ipack(self.g.b) + ipack(self.h.a) + ipack(self.h.b)

    @classmethod
    def unserialize(cls, s):
        rem = s
        nums = []
        while rem and len(nums) < cls.FIELDS:
            unpacked, rem = iunpack(rem)
            nums.append(unpacked)
        if len(nums) != cls.FIELDS:
            return None
        inits = [nums[0],
                 FP2Value(nums[0], nums[1], nums[2]),
                 FP2Value(nums[0], nums[3], nums[4])]
        if len(nums) > 5:
            inits.append(nums[5])
            inits.append(nums[6])
        return cls(*inits)


class BonehPrivateKey(BonehPublicKey):
    """
    A private key for Boneh et al.'s cryptosystem.
    """
    FIELDS = 7

    def __init__(self, p, g, h, n, t1):
        super(BonehPrivateKey, self).__init__(p, g, h)
        self.n = n
        self.t1 = t1

    def serialize(self):
        return super(BonehPrivateKey, self).serialize() + ipack(self.n) + ipack(self.t1)

    def public_key(self):
        return BonehPublicKey(self.p, self.g, self.h)



from __future__ import absolute_import

from binascii import unhexlify
from time import time

from twisted.internet.defer import succeed

from ..identity_formats import FORMATS
from ...attestation.trustchain.community import TrustChainCommunity
from ...attestation.trustchain.listener import BlockListener
from ...peer import Peer


class IdentityCommunity(TrustChainCommunity, BlockListener):

    master_peer = Peer(unhexlify("307e301006072a8648ce3d020106052b81040024036a000400c9104b573ea18b795cb23b1defe6e5b7"
                                 "41afa4b2b5edfe7d211c9342dfb753a22e850fb1bff01d5ca66cfe0b1a845fa3e333d200b6d742151f"
                                 "3e4db3fe8b8508720744c70afe692c73264f789aa36a8c219acebeaa2b6ba652743d6580300fa1d98d"
                                 "96b766dfcd"))

    DB_NAME = 'identity'

    def __init__(self, *args, **kwargs):
        TrustChainCommunity.__init__(self, *args, **kwargs)
        BlockListener.__init__(self)

        self.add_listener(self, [identity_format.encode('utf-8') for identity_format in FORMATS])

        # Dict of hash -> (attribute_name, date, public_key)
        self.known_attestation_hashes = {}

    def add_known_hash(self, attribute_hash, name, public_key, metadata=None):
        """
        We know about this hash+peer combination. Thus we can handle sign requests for it.
        """
        self.known_attestation_hashes[attribute_hash] = (name, time(), public_key, metadata)

    def get_attestation_by_hash(self, attribute_hash):
        blocks = self.persistence.get_all_blocks()
        for block in blocks:
            if block.transaction and block.transaction.get(b"hash", None) == attribute_hash:
                return block
        return None

    def received_block(self, block):
        pass

    def should_sign(self, block):
        transaction = block.transaction
        requested_keys = set(transaction.keys())
        if requested_keys - {b"hash", b"name", b"date", b"metadata"} != set():
            return succeed(False)
        if requested_keys - {b"metadata"} != {b"hash", b"name", b"date"}:
            return succeed(False)
        attribute_hash = transaction[b'hash']
        if attribute_hash not in self.known_attestation_hashes:
            return succeed(False)
        if block.public_key != self.known_attestation_hashes[attribute_hash][2]:
            return succeed(False)
        # Refuse to sign blocks older than 5 minutes
        if time() > self.known_attestation_hashes[attribute_hash][1] + 300:
            return succeed(False)
        if transaction[b'name'] != self.known_attestation_hashes[attribute_hash][0]:
            return succeed(False)
        if (self.known_attestation_hashes[attribute_hash][3]
                and transaction.get(b'metadata', None) != self.known_attestation_hashes[attribute_hash][3]):
            return succeed(False)
        return succeed(True)

    def request_attestation_advertisement(self, peer, attribute_hash, name, block_type="id_metadata", metadata=None):
        """
        Request a peer to sign for our attestation advertisement.
        :param peer: the attestor of our block
        :param attribute_hash: the hash of the attestation
        :param name: the name of the attribute (metadata)
        :param block_type: the type of block (from identity_foromats.py)
        :param metadata: custom additional metadata
        """
        self.sign_block(peer,
                        public_key=peer.public_key.key_to_bin(),
                        block_type=block_type.encode('utf-8'),
                        transaction={
                            b"hash": attribute_hash,
                            b"name": name,
                            b"date": time(),
                            b"metadata": metadata or {}
                        })



import json
from base64 import b64encode

from twisted.internet.defer import inlineCallbacks, returnValue

from .rest_api_peer import InteractiveRestTestPeer
from .rest_peer_communication import string_to_url


class AECommonBehaviorTestPeer(InteractiveRestTestPeer):
    """
    This class implements some auxiliary methods which may be common across a number of test peers with interaction
    facilities
    """

    @inlineCallbacks
    def wait_for_peers(self, dict_param, excluded_peer_mids=None):
        """
        Wait until this peer receives a non-empty list of fellow peers in the network

        :param dict_param: the required parameters by the GET request generator for the peers request type
        :param excluded_peer_mids: A list of peer mids which should not be taken into consideration as valid peers
        :return: a list of currently known peers in the network bar those passed as exceptions
        """
        assert isinstance(excluded_peer_mids, (list, set)) or not excluded_peer_mids, "excluded_peer_mids " \
                                                                                      "must be a list or set or None"

        # Make sure excluded_peer_mids is a set
        if not excluded_peer_mids:
            excluded_peer_mids = set()
        elif isinstance(excluded_peer_mids, list):
            excluded_peer_mids = set(excluded_peer_mids)

        peer_list = yield self._get_style_requests.make_peers(dict_param)
        peer_list = set(peer_list)

        # Keep iterating until peer_list is non-empty
        while not peer_list - excluded_peer_mids:
            yield self.sleep()

            # Forward and wait for the response
            peer_list = yield self._get_style_requests.make_peers(dict_param)
            peer_list = set(peer_list)

        # Return the peer list
        returnValue(list(peer_list - excluded_peer_mids))

    @inlineCallbacks
    def wait_for_attestation_request(self, dict_param):
        """
        Wait until this peer receives a non-empty list of outstanding attestation requests

        :param dict_param: the required parameters by the GET request generator for the outstanding request type
        :return: a list of outstanding attestation requests
        """
        self._logger.info("Attempting to acquire a list of outstanding requests...")
        outstanding_requests = yield self._get_style_requests.make_outstanding(dict_param)

        # Keep iterating until peer_list is non-empty
        while not outstanding_requests:
            self._logger.info("Could not acquire a list of outstanding requests. Will wait 0.1 seconds and retry.")
            yield self.sleep()

            # Forward and wait for the response
            outstanding_requests = yield self._get_style_requests.make_outstanding(dict_param)

        # Return the peer list
        self._logger.info("Have found a non-empty list of outstanding requests. Returning it.")
        returnValue(outstanding_requests)


class RequesterRestTestPeer(AECommonBehaviorTestPeer):
    """
    Simulates the android application
    """

    def __init__(self, port, overlay_classes, get_style_requests, post_style_requests, param_dict,
                 interface='127.0.0.1', memory_dbs=True):
        """
        AndroidTestPeer initializer

        :param port: this peer's port
        :param overlay_classes: the set of overlay classes which should be contained in the peer's IPv8 session object
        :param get_style_requests: GET style request generator
        :param post_style_requests: POST style request generator
        :param param_dict: a dictionary containing the required parameters to communicate with a peer
        :param interface: IP or alias of the peer. Defaults to '127.0.0.1'
        :param memory_dbs: if True, then the DBs of the various overlays / communities are stored in memory; on disk
                           if False
        """
        AECommonBehaviorTestPeer.__init__(self, port, overlay_classes, get_style_requests, post_style_requests,
                                          interface=interface, memory_dbs=memory_dbs)

        self._param_dict = param_dict
        self._param_dict['port'] = port
        self._param_dict['attribute_value'] = string_to_url(b64encode(b'binarydata'), True)
        self._param_dict['metadata'] = b64encode(json.dumps({'psn': '1234567890'}).encode('utf-8')).decode('utf-8')

    @inlineCallbacks
    def run(self):
        # Wait for a short period of time
        yield self.sleep()

        peer_list = yield self.wait_for_peers(self._param_dict)
        for peer in peer_list:
            self._param_dict['mid'] = string_to_url(peer)

            self._logger.info("Sending an attestation request to %s", self._param_dict['mid'])
            yield self._post_style_requests.make_attestation_request(self._param_dict)


class MinimalActivityRestTestPeer(AECommonBehaviorTestPeer):
    """
    Simulates a minimal activity test peer, which only attempts to discover fellow peers then goes inactive
    """

    def __init__(self, port, overlay_classes, get_style_requests, param_dict, interface='127.0.0.1', memory_dbs=True):
        """
        MinimalActivityTestPeer initializer

        :param port: this peer's port
        :param overlay_classes: the set of overlay classes which should be contained in the peer's IPv8 session object
        :param get_style_requests: GET style request generator
        :param param_dict: a dictionary containing the required parameters to communicate with a peer
        :param interface: IP or alias of the peer. Defaults to '127.0.0.1'
        """
        AECommonBehaviorTestPeer.__init__(self, port, overlay_classes, get_style_requests, None,
                                          interface=interface, memory_dbs=memory_dbs)

        self._param_dict = param_dict
        self._param_dict['port'] = port

    @inlineCallbacks
    def run(self):
        # Wait for a short period of time
        yield self.sleep()

        # Await for some fellow peers, then become inactive
        yield self.wait_for_peers(self._param_dict)
